from json import dumps

from fastapi.openapi.utils import get_openapi

from tests.fastapi.app import app


def test_openapi_schema_generation():
    openapi_schema_json_str = dumps(
        get_openapi(
            title=app.title,
            version=app.version,
            openapi_version=app.openapi_version,
            description=app.description,
            routes=app.routes,
        ),
    )

    assert openapi_schema_json_str is not None



import warnings
from datetime import datetime, timedelta, timezone
from random import randint
from typing import List

import pytest

from beanie.odm.utils.init import init_beanie
from tests.odm.models import (
    ADocument,
    BDocument,
    Bicycle,
    Bike,
    BsonRegexDoc,
    Bus,
    Car,
    Doc2NonRoot,
    DocNonRoot,
    DocumentForEncodingTest,
    DocumentForEncodingTestDate,
    DocumentMultiModelOne,
    DocumentMultiModelTwo,
    DocumentTestModel,
    DocumentTestModelFailInspection,
    DocumentTestModelIndexFlagsAnnotated,
    DocumentTestModelWithComplexIndex,
    DocumentTestModelWithCustomCollectionName,
    DocumentTestModelWithIndexFlags,
    DocumentTestModelWithIndexFlagsAliases,
    DocumentTestModelWithLink,
    DocumentTestModelWithSimpleIndex,
    DocumentTestModelWithSoftDelete,
    DocumentToBeLinked,
    DocumentToTestSync,
    DocumentUnion,
    DocumentWithActions,
    DocumentWithActions2,
    DocumentWithBackLink,
    DocumentWithBackLinkForNesting,
    DocumentWithBsonBinaryField,
    DocumentWithBsonEncodersFiledsTypes,
    DocumentWithComplexDictKey,
    DocumentWithCustomFiledsTypes,
    DocumentWithCustomIdInt,
    DocumentWithCustomIdUUID,
    DocumentWithCustomInit,
    DocumentWithDecimalField,
    DocumentWithDeprecatedHiddenField,
    DocumentWithEnumKeysDict,
    DocumentWithExtras,
    DocumentWithHttpUrlField,
    DocumentWithIndexedObjectId,
    DocumentWithIndexMerging1,
    DocumentWithIndexMerging2,
    DocumentWithKeepNullsFalse,
    DocumentWithLink,
    DocumentWithLinkForNesting,
    DocumentWithList,
    DocumentWithListBackLink,
    DocumentWithListLink,
    DocumentWithListOfLinks,
    DocumentWithOptionalBackLink,
    DocumentWithOptionalListBackLink,
    DocumentWithPydanticConfig,
    DocumentWithRevisionTurnedOn,
    DocumentWithRootModelAsAField,
    DocumentWithStringField,
    DocumentWithTextIndexAndLink,
    DocumentWithTimeStampToTestConsistency,
    DocumentWithTurnedOffStateManagement,
    DocumentWithTurnedOnReplaceObjects,
    DocumentWithTurnedOnSavePrevious,
    DocumentWithTurnedOnStateManagement,
    DocumentWithTurnedOnStateManagementWithCustomId,
    DocumentWithValidationOnSave,
    DocWithCallWrapper,
    Door,
    GeoObject,
    House,
    HouseWithRevision,
    InheritedDocumentWithActions,
    LinkDocumentForTextSeacrh,
    Lock,
    LockWithRevision,
    LongSelfLink,
    LoopedLinksA,
    LoopedLinksB,
    NativeRegexDoc,
    Nested,
    Option1,
    Option2,
    Owner,
    PackageElemMatch,
    Region,
    Roof,
    RootDocument,
    Sample,
    SampleLazyParsing,
    SampleWithMutableObjects,
    SelfLinked,
    StateAndDecimalFieldModel,
    SubDocument,
    UsersAddresses,
    Vehicle,
    Window,
    WindowWithRevision,
    WindowWithValidationOnSave,
    Yard,
    YardWithRevision,
)
from tests.odm.views import ViewForTest, ViewForTestWithLink

TESTING_MODELS = [
    DocumentWithExtras,
    DocumentWithPydanticConfig,
    DocumentTestModel,
    DocumentTestModelWithSoftDelete,
    DocumentTestModelWithLink,
    DocumentTestModelWithCustomCollectionName,
    DocumentTestModelWithSimpleIndex,
    DocumentTestModelWithIndexFlags,
    DocumentTestModelWithIndexFlagsAliases,
    DocumentTestModelIndexFlagsAnnotated,
    DocumentTestModelWithComplexIndex,
    DocumentTestModelFailInspection,
    DocumentWithBsonEncodersFiledsTypes,
    DocumentWithCustomFiledsTypes,
    DocumentWithCustomIdUUID,
    DocumentWithCustomIdInt,
    Sample,
    DocumentWithActions,
    DocumentWithTurnedOnStateManagement,
    DocumentWithTurnedOnReplaceObjects,
    DocumentWithTurnedOnSavePrevious,
    DocumentWithTurnedOffStateManagement,
    DocumentWithValidationOnSave,
    DocumentWithRevisionTurnedOn,
    DocumentWithHttpUrlField,
    House,
    Window,
    WindowWithValidationOnSave,
    Door,
    Roof,
    Yard,
    Lock,
    InheritedDocumentWithActions,
    DocumentForEncodingTest,
    DocumentForEncodingTestDate,
    DocumentWithStringField,
    ViewForTest,
    ViewForTestWithLink,
    DocumentMultiModelOne,
    DocumentMultiModelTwo,
    DocumentUnion,
    HouseWithRevision,
    WindowWithRevision,
    LockWithRevision,
    YardWithRevision,
    DocumentWithActions2,
    Vehicle,
    Bicycle,
    Bike,
    Car,
    Bus,
    Owner,
    SampleWithMutableObjects,
    DocNonRoot,
    Doc2NonRoot,
    SampleLazyParsing,
    RootDocument,
    ADocument,
    BDocument,
    StateAndDecimalFieldModel,
    Region,
    UsersAddresses,
    SelfLinked,
    LoopedLinksA,
    LoopedLinksB,
    DocumentWithTurnedOnStateManagementWithCustomId,
    DocumentWithDecimalField,
    DocumentWithKeepNullsFalse,
    PackageElemMatch,
    DocumentWithLink,
    DocumentWithBackLink,
    DocumentWithListLink,
    DocumentWithListBackLink,
    DocumentWithListOfLinks,
    DocumentToBeLinked,
    DocumentWithTimeStampToTestConsistency,
    DocumentWithIndexMerging1,
    DocumentWithIndexMerging2,
    DocumentWithCustomInit,
    DocumentWithTextIndexAndLink,
    LinkDocumentForTextSeacrh,
    DocumentWithList,
    DocumentWithBsonBinaryField,
    DocumentWithRootModelAsAField,
    DocWithCallWrapper,
    DocumentWithOptionalBackLink,
    DocumentWithOptionalListBackLink,
    DocumentWithComplexDictKey,
    DocumentWithIndexedObjectId,
    DocumentToTestSync,
    DocumentWithLinkForNesting,
    DocumentWithBackLinkForNesting,
    DocumentWithEnumKeysDict,
    LongSelfLink,
    BsonRegexDoc,
    NativeRegexDoc,
]


@pytest.fixture
def point():
    return {
        "longitude": 13.404954,
        "latitude": 52.520008,
    }


@pytest.fixture
async def preset_documents(point):
    docs = []
    for i in range(10):
        timestamp = datetime.now(tz=timezone.utc) - timedelta(days=i)
        integer_1: int = i // 3
        integer_2: int = i // 2
        float_num = integer_1 + 0.3
        string: str = f"test_{integer_1}"
        option_1 = Option1(s="TEST")
        option_2 = Option2(f=3.14)
        union = option_1 if i % 2 else option_2
        optional = option_2 if not i % 3 else None
        geo = GeoObject(
            coordinates=[
                point["longitude"] + i / 10,
                point["latitude"] + i / 10,
            ]
        )
        nested = Nested(
            integer=integer_2,
            option_1=option_1,
            union=union,
            optional=optional,
        )

        const = "TEST"

        sample = Sample(
            timestamp=timestamp,
            increment=i,
            integer=integer_1,
            float_num=float_num,
            string=string,
            nested=nested,
            optional=optional,
            union=union,
            geo=geo,
            const=const,
        )

        docs.append(sample)
    await Sample.insert_many(documents=docs)


@pytest.fixture()
def sample_doc_not_saved(point):
    nested = Nested(
        integer=0,
        option_1=Option1(s="TEST"),
        union=Option1(s="TEST"),
        optional=None,
    )
    geo = GeoObject(
        coordinates=[
            point["longitude"],
            point["latitude"],
        ]
    )
    return Sample(
        timestamp=datetime.now(tz=timezone.utc),
        increment=0,
        integer=0,
        float_num=0,
        string="TEST_NOT_SAVED",
        nested=nested,
        optional=None,
        union=Option1(s="TEST"),
        geo=geo,
    )


@pytest.fixture()
async def session(cli):
    s = await cli.start_session()
    yield s
    await s.end_session()


@pytest.fixture()
async def deprecated_init_beanie(db):
    for model in TESTING_MODELS:  # crude clear from init
        await model.get_motor_collection().drop()
        await model.get_motor_collection().drop_indexes()
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")

        await init_beanie(
            database=db,
            document_models=[DocumentWithDeprecatedHiddenField],
        )
        assert len(w) == 1
        assert issubclass(w[-1].category, DeprecationWarning)
        assert (
            "DocumentWithDeprecatedHiddenField: 'hidden=True' is deprecated, please use 'exclude=True'"
            in str(w[-1].message)
        )


@pytest.fixture(autouse=True)
async def init(db):
    await init_beanie(
        database=db,
        document_models=TESTING_MODELS,
    )
    try:
        yield None
    finally:
        for model in TESTING_MODELS:
            await model.get_motor_collection().drop()
            await model.get_motor_collection().drop_indexes()


@pytest.fixture
def document_not_inserted():
    return DocumentTestModel(
        test_int=42,
        test_list=[SubDocument(test_str="foo"), SubDocument(test_str="bar")],
        test_doc=SubDocument(test_str="foobar"),
        test_str="kipasa",
    )


@pytest.fixture
def documents_not_inserted():
    def generate_documents(
        number: int, test_str: str = None, random: bool = False
    ) -> List[DocumentTestModel]:
        return [
            DocumentTestModel(
                test_int=randint(0, 1000000) if random else i,
                test_list=[
                    SubDocument(test_str="foo"),
                    SubDocument(test_str="bar"),
                ],
                test_doc=SubDocument(test_str="foobar"),
                test_str="kipasa" if test_str is None else test_str,
            )
            for i in range(number)
        ]

    return generate_documents


@pytest.fixture
def document_soft_delete_not_inserted():
    return DocumentTestModelWithSoftDelete(
        test_int=randint(0, 1000000),
        test_str="kipasa",
    )


@pytest.fixture
def documents_soft_delete_not_inserted():
    docs = []
    for i in range(3):
        docs.append(
            DocumentTestModelWithSoftDelete(
                test_int=randint(0, 1000000),
                test_str="kipasa",
            )
        )
    return docs


@pytest.fixture
async def document(document_not_inserted) -> DocumentTestModel:
    return await document_not_inserted.insert()


@pytest.fixture
def documents(documents_not_inserted):
    async def generate_documents(
        number: int, test_str: str = None, random: bool = False
    ):
        result = await DocumentTestModel.insert_many(
            documents_not_inserted(number, test_str, random)
        )
        return result.inserted_ids

    return generate_documents


@pytest.fixture
def documents_with_links(documents):
    async def generate_documents():
        await documents(15)
        results = await DocumentTestModel.all().to_list()
        for document in results:
            await DocumentTestModelWithLink(test_link=document).insert()

    return generate_documents



from abc import abstractmethod
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Mapping,
    Optional,
    Type,
    Union,
)

from motor.motor_asyncio import AsyncIOMotorClientSession
from pymongo import ReturnDocument
from pymongo import UpdateMany as UpdateManyPyMongo
from pymongo import UpdateOne as UpdateOnePyMongo
from pymongo.results import InsertOneResult, UpdateResult

from beanie.odm.bulk import BulkWriter
from beanie.odm.interfaces.clone import CloneInterface
from beanie.odm.interfaces.session import SessionMethods
from beanie.odm.interfaces.update import (
    UpdateMethods,
)
from beanie.odm.operators.update import BaseUpdateOperator
from beanie.odm.operators.update.general import SetRevisionId
from beanie.odm.utils.encoder import Encoder
from beanie.odm.utils.parsing import parse_obj

if TYPE_CHECKING:
    from beanie.odm.documents import DocType


class UpdateResponse(str, Enum):
    UPDATE_RESULT = "UPDATE_RESULT"  # PyMongo update result
    OLD_DOCUMENT = "OLD_DOCUMENT"  # Original document
    NEW_DOCUMENT = "NEW_DOCUMENT"  # Updated document


class UpdateQuery(UpdateMethods, SessionMethods, CloneInterface):
    """
    Update Query base class
    """

    def __init__(
        self,
        document_model: Type["DocType"],
        find_query: Mapping[str, Any],
    ):
        self.document_model = document_model
        self.find_query = find_query
        self.update_expressions: List[Mapping[str, Any]] = []
        self.session = None
        self.is_upsert = False
        self.upsert_insert_doc: Optional["DocType"] = None
        self.encoders: Dict[Any, Callable[[Any], Any]] = {}
        self.bulk_writer: Optional[BulkWriter] = None
        self.encoders = self.document_model.get_settings().bson_encoders
        self.pymongo_kwargs: Dict[str, Any] = {}

    @property
    def update_query(self) -> Dict[str, Any]:
        query: Union[Dict[str, Any], List[Dict[str, Any]], None] = None
        for expression in self.update_expressions:
            if isinstance(expression, BaseUpdateOperator):
                if query is None:
                    query = {}
                if isinstance(query, list):
                    raise TypeError("Wrong expression type")
                query.update(expression.query)
            elif isinstance(expression, dict):
                if query is None:
                    query = {}
                if isinstance(query, list):
                    raise TypeError("Wrong expression type")
                query.update(expression)
            elif isinstance(expression, SetRevisionId):
                if query is None:
                    query = {}
                if isinstance(query, list):
                    raise TypeError("Wrong expression type")
                set_query = query.get("$set", {})
                set_query.update(expression.query.get("$set", {}))
                query["$set"] = set_query
            elif isinstance(expression, list):
                if query is None:
                    query = []
                if isinstance(query, dict):
                    raise TypeError("Wrong expression type")
                query.extend(expression)
            else:
                raise TypeError("Wrong expression type")
        return Encoder(custom_encoders=self.encoders).encode(query)

    @abstractmethod
    async def _update(self) -> UpdateResult: ...


class UpdateMany(UpdateQuery):
    """
    Update Many query class
    """

    def update(
        self,
        *args: Mapping[str, Any],
        session: Optional[AsyncIOMotorClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        **pymongo_kwargs: Any,
    ) -> "UpdateQuery":
        """
        Provide modifications to the update query.

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param session: Optional[AsyncIOMotorClientSession]
        :param bulk_writer: Optional[BulkWriter]
        :param pymongo_kwargs: pymongo native parameters for update operation
        :return: UpdateMany query
        """
        self.set_session(session=session)
        self.update_expressions += args
        if bulk_writer:
            self.bulk_writer = bulk_writer
        self.pymongo_kwargs.update(pymongo_kwargs)
        return self

    def upsert(
        self,
        *args: Mapping[str, Any],
        on_insert: "DocType",
        session: Optional[AsyncIOMotorClientSession] = None,
        **pymongo_kwargs: Any,
    ) -> "UpdateQuery":
        """
        Provide modifications to the upsert query.

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param on_insert: DocType - document to insert if there is no matched
        document in the collection
        :param session: Optional[AsyncIOMotorClientSession]
        :param **pymongo_kwargs: pymongo native parameters for update operation
        :return: UpdateMany query
        """
        self.upsert_insert_doc = on_insert  # type: ignore
        self.update(*args, session=session, **pymongo_kwargs)
        return self

    def update_many(
        self,
        *args: Mapping[str, Any],
        session: Optional[AsyncIOMotorClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        **pymongo_kwargs: Any,
    ):
        """
        Provide modifications to the update query

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param session: Optional[AsyncIOMotorClientSession]
        :param bulk_writer: "BulkWriter" - Beanie bulk writer
        :param pymongo_kwargs: pymongo native parameters for update operation
        :return: UpdateMany query
        """
        return self.update(
            *args, session=session, bulk_writer=bulk_writer, **pymongo_kwargs
        )

    async def _update(self):
        if self.bulk_writer is None:
            return (
                await self.document_model.get_motor_collection().update_many(
                    self.find_query,
                    self.update_query,
                    session=self.session,
                    **self.pymongo_kwargs,
                )
            )
        else:
            self.bulk_writer.add_operation(
                self.document_model,
                UpdateManyPyMongo(
                    self.find_query, self.update_query, **self.pymongo_kwargs
                ),
            )

    def __await__(
        self,
    ) -> Generator[
        Any, None, Union[UpdateResult, InsertOneResult, Optional["DocType"]]
    ]:
        """
        Run the query
        :return:
        """

        update_result = yield from self._update().__await__()
        if self.upsert_insert_doc is None:
            return update_result

        if update_result is not None and update_result.matched_count == 0:
            return (
                yield from self.document_model.insert_one(
                    document=self.upsert_insert_doc,
                    session=self.session,
                    bulk_writer=self.bulk_writer,
                ).__await__()
            )

        return update_result


class UpdateOne(UpdateQuery):
    """
    Update One query class
    """

    def __init__(self, *args: Any, **kwargs: Any):
        super(UpdateOne, self).__init__(*args, **kwargs)
        self.response_type = UpdateResponse.UPDATE_RESULT

    def update(
        self,
        *args: Mapping[str, Any],
        session: Optional[AsyncIOMotorClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        response_type: Optional[UpdateResponse] = None,
        **pymongo_kwargs: Any,
    ) -> "UpdateQuery":
        """
        Provide modifications to the update query.

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param session: Optional[AsyncIOMotorClientSession]
        :param bulk_writer: Optional[BulkWriter]
        :param response_type: UpdateResponse
        :param pymongo_kwargs: pymongo native parameters for update operation
        :return: UpdateMany query
        """
        self.set_session(session=session)
        self.update_expressions += args
        if response_type is not None:
            self.response_type = response_type
        if bulk_writer:
            self.bulk_writer = bulk_writer
        self.pymongo_kwargs.update(pymongo_kwargs)
        return self

    def upsert(
        self,
        *args: Mapping[str, Any],
        on_insert: "DocType",
        session: Optional[AsyncIOMotorClientSession] = None,
        response_type: Optional[UpdateResponse] = None,
        **pymongo_kwargs: Any,
    ) -> "UpdateQuery":
        """
        Provide modifications to the upsert query.

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param on_insert: DocType - document to insert if there is no matched
        document in the collection
        :param session: Optional[AsyncIOMotorClientSession]
        :param response_type: Optional[UpdateResponse]
        :param pymongo_kwargs: pymongo native parameters for update operation
        :return: UpdateMany query
        """
        self.upsert_insert_doc = on_insert  # type: ignore
        self.update(
            *args,
            response_type=response_type,
            session=session,
            **pymongo_kwargs,
        )
        return self

    def update_one(
        self,
        *args: Mapping[str, Any],
        session: Optional[AsyncIOMotorClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        response_type: Optional[UpdateResponse] = None,
        **pymongo_kwargs: Any,
    ):
        """
        Provide modifications to the update query. The same as `update()`

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param session: Optional[AsyncIOMotorClientSession]
        :param bulk_writer: "BulkWriter" - Beanie bulk writer
        :param response_type: Optional[UpdateResponse]
        :param pymongo_kwargs: pymongo native parameters for update operation
        :return: UpdateMany query
        """
        return self.update(
            *args,
            session=session,
            bulk_writer=bulk_writer,
            response_type=response_type,
            **pymongo_kwargs,
        )

    async def _update(self):
        if not self.bulk_writer:
            if self.response_type == UpdateResponse.UPDATE_RESULT:
                return await self.document_model.get_motor_collection().update_one(
                    self.find_query,
                    self.update_query,
                    session=self.session,
                    **self.pymongo_kwargs,
                )
            else:
                result = await self.document_model.get_motor_collection().find_one_and_update(
                    self.find_query,
                    self.update_query,
                    session=self.session,
                    return_document=(
                        ReturnDocument.BEFORE
                        if self.response_type == UpdateResponse.OLD_DOCUMENT
                        else ReturnDocument.AFTER
                    ),
                    **self.pymongo_kwargs,
                )
                if result is not None:
                    result = parse_obj(self.document_model, result)
                return result
        else:
            self.bulk_writer.add_operation(
                self.document_model,
                UpdateOnePyMongo(
                    self.find_query, self.update_query, **self.pymongo_kwargs
                ),
            )

    def __await__(
        self,
    ) -> Generator[
        Any, None, Union[UpdateResult, InsertOneResult, Optional["DocType"]]
    ]:
        """
        Run the query
        :return:
        """
        update_result = yield from self._update().__await__()
        if self.upsert_insert_doc is None:
            return update_result

        if (
            self.response_type == UpdateResponse.UPDATE_RESULT
            and update_result is not None
            and update_result.matched_count == 0
        ) or (
            self.response_type != UpdateResponse.UPDATE_RESULT
            and update_result is None
        ):
            return (
                yield from self.document_model.insert_one(
                    document=self.upsert_insert_doc,
                    session=self.session,
                    bulk_writer=self.bulk_writer,
                ).__await__()
            )

        return update_result



from tests.odm.models import DocumentTestModel


async def test_count(documents):
    await documents(4, "uno", True)
    c = await DocumentTestModel.count()
    assert c == 4


async def test_count_with_filter_query(documents):
    await documents(4, "uno", True)
    await documents(2, "dos", True)
    await documents(1, "cuatro", True)
    c = await DocumentTestModel.find_many({"test_str": "dos"}).count()
    assert c == 2


async def test_count_with_limit(documents):
    await documents(5, "five", True)
    c = await DocumentTestModel.find_all().limit(1).count()
    assert c == 1
    d = await DocumentTestModel.find_all().count()
    assert d == 5



import asyncio
import warnings
from datetime import datetime, timezone
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Coroutine,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from uuid import UUID, uuid4

from bson import DBRef, ObjectId
from lazy_model import LazyModel
from motor.motor_asyncio import AsyncIOMotorClientSession
from pydantic import (
    ConfigDict,
    Field,
    PrivateAttr,
    ValidationError,
)
from pydantic.class_validators import root_validator
from pydantic.main import BaseModel
from pymongo import InsertOne
from pymongo.errors import DuplicateKeyError
from pymongo.results import (
    DeleteResult,
    InsertManyResult,
)
from typing_extensions import Concatenate, ParamSpec, Self, TypeAlias

from beanie.exceptions import (
    CollectionWasNotInitialized,
    DocumentNotFound,
    DocumentWasNotSaved,
    NotSupported,
    ReplaceError,
    RevisionIdWasChanged,
)
from beanie.odm.actions import (
    ActionDirections,
    EventTypes,
    wrap_with_actions,
)
from beanie.odm.bulk import BulkWriter
from beanie.odm.cache import LRUCache
from beanie.odm.enums import SortDirection
from beanie.odm.fields import (
    BackLink,
    DeleteRules,
    ExpressionField,
    Link,
    LinkInfo,
    LinkTypes,
    PydanticObjectId,
    WriteRules,
)
from beanie.odm.interfaces.aggregate import AggregateInterface
from beanie.odm.interfaces.detector import ModelType
from beanie.odm.interfaces.find import FindInterface
from beanie.odm.interfaces.getters import OtherGettersInterface
from beanie.odm.interfaces.inheritance import InheritanceInterface
from beanie.odm.interfaces.setters import SettersInterface
from beanie.odm.models import (
    InspectionError,
    InspectionResult,
    InspectionStatuses,
)
from beanie.odm.operators.find.comparison import In
from beanie.odm.operators.update.general import (
    CurrentDate,
    Inc,
    SetRevisionId,
    Unset,
)
from beanie.odm.operators.update.general import (
    Set as SetOperator,
)
from beanie.odm.queries.find import FindMany, FindOne
from beanie.odm.queries.update import UpdateMany, UpdateResponse
from beanie.odm.settings.document import DocumentSettings
from beanie.odm.utils.dump import get_dict, get_top_level_nones
from beanie.odm.utils.parsing import apply_changes, merge_models
from beanie.odm.utils.pydantic import (
    IS_PYDANTIC_V2,
    get_extra_field_info,
    get_field_type,
    get_model_dump,
    get_model_fields,
    parse_model,
    parse_object_as,
)
from beanie.odm.utils.self_validation import validate_self_before
from beanie.odm.utils.state import (
    previous_saved_state_needed,
    save_state_after,
    saved_state_needed,
)
from beanie.odm.utils.typing import extract_id_class

if IS_PYDANTIC_V2:
    from pydantic import model_validator

if TYPE_CHECKING:
    from beanie.odm.views import View

FindType = TypeVar("FindType", bound=Union["Document", "View"])
DocType = TypeVar("DocType", bound="Document")
P = ParamSpec("P")
R = TypeVar("R")
# can describe both sync and async, where R itself is a coroutine
AnyDocMethod: TypeAlias = Callable[Concatenate[DocType, P], R]
# describes only async
AsyncDocMethod: TypeAlias = Callable[
    Concatenate[DocType, P], Coroutine[Any, Any, R]
]
DocumentProjectionType = TypeVar("DocumentProjectionType", bound=BaseModel)


def json_schema_extra(schema: Dict[str, Any], model: Type["Document"]) -> None:
    # remove excluded fields from the json schema
    properties = schema.get("properties")
    if not properties:
        return
    for k, field in get_model_fields(model).items():
        k = field.alias or k
        if k not in properties:
            continue
        field_info = field if IS_PYDANTIC_V2 else field.field_info
        if field_info.exclude:
            del properties[k]


def document_alias_generator(s: str) -> str:
    if s == "id":
        return "_id"
    return s


class MergeStrategy(str, Enum):
    local = "local"
    remote = "remote"


class Document(
    LazyModel,
    SettersInterface,
    InheritanceInterface,
    FindInterface,
    AggregateInterface,
    OtherGettersInterface,
):
    """
    Document Mapping class.

    Fields:

    - `id` - MongoDB document ObjectID "_id" field.
    Mapped to the PydanticObjectId class
    """

    if IS_PYDANTIC_V2:
        model_config = ConfigDict(
            json_schema_extra=json_schema_extra,
            populate_by_name=True,
            alias_generator=document_alias_generator,
        )
    else:

        class Config:
            json_encoders = {ObjectId: str}
            allow_population_by_field_name = True
            fields = {"id": "_id"}
            schema_extra = staticmethod(json_schema_extra)

    id: Optional[PydanticObjectId] = Field(
        default=None, description="MongoDB document ObjectID"
    )

    # State
    revision_id: Optional[UUID] = Field(default=None, exclude=True)
    _saved_state: Optional[Dict[str, Any]] = PrivateAttr(default=None)
    _previous_saved_state: Optional[Dict[str, Any]] = PrivateAttr(default=None)

    # Relations
    _link_fields: ClassVar[Optional[Dict[str, LinkInfo]]] = None

    # Cache
    _cache: ClassVar[Optional[LRUCache]] = None

    # Settings
    _document_settings: ClassVar[Optional[DocumentSettings]] = None

    # Database
    _database_major_version: ClassVar[int] = 4

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super(Document, self).__init__(*args, **kwargs)
        self.get_motor_collection()

    @classmethod
    def _fill_back_refs(cls, values):
        if cls._link_fields:
            for field_name, link_info in cls._link_fields.items():
                if (
                    link_info.link_type
                    in [LinkTypes.BACK_DIRECT, LinkTypes.OPTIONAL_BACK_DIRECT]
                    and field_name not in values
                ):
                    values[field_name] = BackLink[link_info.document_class](
                        link_info.document_class
                    )
                if (
                    link_info.link_type
                    in [LinkTypes.BACK_LIST, LinkTypes.OPTIONAL_BACK_LIST]
                    and field_name not in values
                ):
                    values[field_name] = [
                        BackLink[link_info.document_class](
                            link_info.document_class
                        )
                    ]
        return values

    if IS_PYDANTIC_V2:

        @model_validator(mode="before")
        def fill_back_refs(cls, values):
            return cls._fill_back_refs(values)

    else:

        @root_validator(pre=True)
        def fill_back_refs(cls, values):
            return cls._fill_back_refs(values)

    @classmethod
    async def get(
        cls: Type["DocType"],
        document_id: Any,
        session: Optional[AsyncIOMotorClientSession] = None,
        ignore_cache: bool = False,
        fetch_links: bool = False,
        with_children: bool = False,
        nesting_depth: Optional[int] = None,
        nesting_depths_per_field: Optional[Dict[str, int]] = None,
        **pymongo_kwargs: Any,
    ) -> Optional["DocType"]:
        """
        Get document by id, returns None if document does not exist

        :param document_id: PydanticObjectId - document id
        :param session: Optional[AsyncIOMotorClientSession] - motor session
        :param ignore_cache: bool - ignore cache (if it is turned on)
        :param **pymongo_kwargs: pymongo native parameters for find operation
        :return: Union["Document", None]
        """
        if not isinstance(
            document_id,
            extract_id_class(get_field_type(get_model_fields(cls)["id"])),
        ):
            document_id = parse_object_as(
                get_field_type(get_model_fields(cls)["id"]), document_id
            )

        return await cls.find_one(
            {"_id": document_id},
            session=session,
            ignore_cache=ignore_cache,
            fetch_links=fetch_links,
            with_children=with_children,
            nesting_depth=nesting_depth,
            nesting_depths_per_field=nesting_depths_per_field,
            **pymongo_kwargs,
        )

    async def sync(self, merge_strategy: MergeStrategy = MergeStrategy.remote):
        """
        Sync the document with the database

        :param merge_strategy: MergeStrategy - how to merge the document
        :return: None
        """
        if (
            merge_strategy == MergeStrategy.local
            and self.get_settings().use_state_management is False
        ):
            raise ValueError(
                "State management must be turned on to use local merge strategy"
            )
        if self.id is None:
            raise DocumentWasNotSaved
        document = await self.find_one({"_id": self.id})
        if document is None:
            raise DocumentNotFound

        if merge_strategy == MergeStrategy.local:
            original_changes = self.get_changes()
            new_state = document.get_saved_state()
            if new_state is None:
                raise DocumentWasNotSaved
            changes_to_apply = self._collect_updates(
                new_state, original_changes
            )
            merge_models(self, document)
            apply_changes(changes_to_apply, self)
        elif merge_strategy == MergeStrategy.remote:
            merge_models(self, document)
        else:
            raise ValueError("Invalid merge strategy")

    @wrap_with_actions(EventTypes.INSERT)
    @save_state_after
    @validate_self_before
    async def insert(
        self: Self,
        *,
        link_rule: WriteRules = WriteRules.DO_NOTHING,
        session: Optional[AsyncIOMotorClientSession] = None,
        skip_actions: Optional[List[Union[ActionDirections, str]]] = None,
    ) -> Self:
        """
        Insert the document (self) to the collection
        :return: self
        """
        if self.get_settings().use_revision:
            self.revision_id = uuid4()
        if link_rule == WriteRules.WRITE:
            link_fields = self.get_link_fields()
            if link_fields is not None:
                for field_info in link_fields.values():
                    value = getattr(self, field_info.field_name)
                    if field_info.link_type in [
                        LinkTypes.DIRECT,
                        LinkTypes.OPTIONAL_DIRECT,
                    ]:
                        if isinstance(value, Document):
                            await value.save(
                                link_rule=WriteRules.WRITE, session=session
                            )
                    if field_info.link_type in [
                        LinkTypes.LIST,
                        LinkTypes.OPTIONAL_LIST,
                    ]:
                        if isinstance(value, List):
                            await asyncio.gather(
                                *[
                                    obj.save(
                                        link_rule=WriteRules.WRITE,
                                        session=session,
                                    )
                                    for obj in value
                                    if isinstance(obj, Document)
                                ]
                            )
        result = await self.get_motor_collection().insert_one(
            get_dict(
                self, to_db=True, keep_nulls=self.get_settings().keep_nulls
            ),
            session=session,
        )
        new_id = result.inserted_id
        if not isinstance(
            new_id,
            extract_id_class(get_field_type(get_model_fields(self)["id"])),
        ):
            new_id = parse_object_as(
                get_field_type(get_model_fields(self)["id"]), new_id
            )
        self.id = new_id
        return self

    async def create(
        self: Self,
        session: Optional[AsyncIOMotorClientSession] = None,
    ) -> Self:
        """
        The same as self.insert()
        :return: self
        """
        return await self.insert(session=session)

    @classmethod
    async def insert_one(
        cls: Type[DocType],
        document: DocType,
        session: Optional[AsyncIOMotorClientSession] = None,
        bulk_writer: Optional["BulkWriter"] = None,
        link_rule: WriteRules = WriteRules.DO_NOTHING,
    ) -> Optional[DocType]:
        """
        Insert one document to the collection
        :param document: Document - document to insert
        :param session: AsyncIOMotorClientSession - motor session
        :param bulk_writer: "BulkWriter" - Beanie bulk writer
        :param link_rule: InsertRules - hot to manage link fields
        :return: DocType
        """
        if not isinstance(document, cls):
            raise TypeError(
                "Inserting document must be of the original document class"
            )
        if bulk_writer is None:
            return await document.insert(link_rule=link_rule, session=session)
        else:
            if link_rule == WriteRules.WRITE:
                raise NotSupported(
                    "Cascade insert with bulk writing not supported"
                )
            bulk_writer.add_operation(
                type(document),
                InsertOne(
                    get_dict(
                        document,
                        to_db=True,
                        keep_nulls=document.get_settings().keep_nulls,
                    )
                ),
            )
            return None

    @classmethod
    async def insert_many(
        cls: Type[DocType],
        documents: Iterable[DocType],
        session: Optional[AsyncIOMotorClientSession] = None,
        link_rule: WriteRules = WriteRules.DO_NOTHING,
        **pymongo_kwargs: Any,
    ) -> InsertManyResult:
        """
        Insert many documents to the collection

        :param documents:  List["Document"] - documents to insert
        :param session: AsyncIOMotorClientSession - motor session
        :param link_rule: InsertRules - how to manage link fields
        :return: InsertManyResult
        """
        if link_rule == WriteRules.WRITE:
            raise NotSupported(
                "Cascade insert not supported for insert many method"
            )
        documents_list = [
            get_dict(
                document,
                to_db=True,
                keep_nulls=document.get_settings().keep_nulls,
            )
            for document in documents
        ]
        return await cls.get_motor_collection().insert_many(
            documents_list, session=session, **pymongo_kwargs
        )

    @wrap_with_actions(EventTypes.REPLACE)
    @save_state_after
    @validate_self_before
    async def replace(
        self: Self,
        ignore_revision: bool = False,
        session: Optional[AsyncIOMotorClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        link_rule: WriteRules = WriteRules.DO_NOTHING,
        skip_actions: Optional[List[Union[ActionDirections, str]]] = None,
    ) -> Self:
        """
        Fully update the document in the database

        :param session: Optional[AsyncIOMotorClientSession] - motor session.
        :param ignore_revision: bool - do force replace.
            Used when revision based protection is turned on.
        :param bulk_writer: "BulkWriter" - Beanie bulk writer
        :return: self
        """
        if self.id is None:
            raise ValueError("Document must have an id")

        if bulk_writer is not None and link_rule != WriteRules.DO_NOTHING:
            raise NotSupported

        if link_rule == WriteRules.WRITE:
            link_fields = self.get_link_fields()
            if link_fields is not None:
                for field_info in link_fields.values():
                    value = getattr(self, field_info.field_name)
                    if field_info.link_type in [
                        LinkTypes.DIRECT,
                        LinkTypes.OPTIONAL_DIRECT,
                        LinkTypes.BACK_DIRECT,
                        LinkTypes.OPTIONAL_BACK_DIRECT,
                    ]:
                        if isinstance(value, Document):
                            await value.replace(
                                link_rule=link_rule,
                                bulk_writer=bulk_writer,
                                ignore_revision=ignore_revision,
                                session=session,
                            )
                    if field_info.link_type in [
                        LinkTypes.LIST,
                        LinkTypes.OPTIONAL_LIST,
                        LinkTypes.BACK_LIST,
                        LinkTypes.OPTIONAL_BACK_LIST,
                    ]:
                        if isinstance(value, List):
                            await asyncio.gather(
                                *[
                                    obj.replace(
                                        link_rule=link_rule,
                                        bulk_writer=bulk_writer,
                                        ignore_revision=ignore_revision,
                                        session=session,
                                    )
                                    for obj in value
                                    if isinstance(obj, Document)
                                ]
                            )

        use_revision_id = self.get_settings().use_revision
        find_query: Dict[str, Any] = {"_id": self.id}

        if use_revision_id and not ignore_revision:
            find_query["revision_id"] = self.revision_id
            self.revision_id = uuid4()
        try:
            await self.find_one(find_query).replace_one(
                self,
                session=session,
                bulk_writer=bulk_writer,
            )
        except DocumentNotFound:
            if use_revision_id and not ignore_revision:
                raise RevisionIdWasChanged
            else:
                raise DocumentNotFound
        return self

    @wrap_with_actions(EventTypes.SAVE)
    @save_state_after
    @validate_self_before
    async def save(
        self: Self,
        session: Optional[AsyncIOMotorClientSession] = None,
        link_rule: WriteRules = WriteRules.DO_NOTHING,
        ignore_revision: bool = False,
        **kwargs: Any,
    ) -> Self:
        """
        Update an existing model in the database or
        insert it if it does not yet exist.

        :param session: Optional[AsyncIOMotorClientSession] - motor session.
        :param link_rule: WriteRules - rules how to deal with links on writing
        :param ignore_revision: bool - do force save.
        :return: self
        """
        if link_rule == WriteRules.WRITE:
            link_fields = self.get_link_fields()
            if link_fields is not None:
                for field_info in link_fields.values():
                    value = getattr(self, field_info.field_name)
                    if field_info.link_type in [
                        LinkTypes.DIRECT,
                        LinkTypes.OPTIONAL_DIRECT,
                        LinkTypes.BACK_DIRECT,
                        LinkTypes.OPTIONAL_BACK_DIRECT,
                    ]:
                        if isinstance(value, Document):
                            await value.save(
                                link_rule=link_rule, session=session
                            )
                    if field_info.link_type in [
                        LinkTypes.LIST,
                        LinkTypes.OPTIONAL_LIST,
                        LinkTypes.BACK_LIST,
                        LinkTypes.OPTIONAL_BACK_LIST,
                    ]:
                        if isinstance(value, List):
                            await asyncio.gather(
                                *[
                                    obj.save(
                                        link_rule=link_rule, session=session
                                    )
                                    for obj in value
                                    if isinstance(obj, Document)
                                ]
                            )

        if self.get_settings().keep_nulls is False:
            return await self.update(
                SetOperator(
                    get_dict(
                        self,
                        to_db=True,
                        keep_nulls=self.get_settings().keep_nulls,
                    )
                ),
                Unset(get_top_level_nones(self)),
                session=session,
                ignore_revision=ignore_revision,
                upsert=True,
                **kwargs,
            )
        else:
            return await self.update(
                SetOperator(
                    get_dict(
                        self,
                        to_db=True,
                        keep_nulls=self.get_settings().keep_nulls,
                    )
                ),
                session=session,
                ignore_revision=ignore_revision,
                upsert=True,
                **kwargs,
            )

    @saved_state_needed
    @wrap_with_actions(EventTypes.SAVE_CHANGES)
    @validate_self_before
    async def save_changes(
        self: Self,
        ignore_revision: bool = False,
        session: Optional[AsyncIOMotorClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        skip_actions: Optional[List[Union[ActionDirections, str]]] = None,
    ) -> Optional[Self]:
        """
        Save changes.
        State management usage must be turned on

        :param ignore_revision: bool - ignore revision id, if revision is turned on
        :param bulk_writer: "BulkWriter" - Beanie bulk writer
        :return: Optional[self]
        """
        if not self.is_changed:
            return None
        changes = self.get_changes()
        if self.get_settings().keep_nulls is False:
            return await self.update(
                SetOperator(changes),
                Unset(get_top_level_nones(self)),
                ignore_revision=ignore_revision,
                session=session,
                bulk_writer=bulk_writer,
            )
        else:
            return await self.set(
                changes,
                ignore_revision=ignore_revision,
                session=session,
                bulk_writer=bulk_writer,
            )

    @classmethod
    async def replace_many(
        cls: Type[DocType],
        documents: List[DocType],
        session: Optional[AsyncIOMotorClientSession] = None,
    ) -> None:
        """
        Replace list of documents

        :param documents: List["Document"]
        :param session: Optional[AsyncIOMotorClientSession] - motor session.
        :return: None
        """
        ids_list = [document.id for document in documents]
        if await cls.find(In(cls.id, ids_list)).count() != len(ids_list):
            raise ReplaceError(
                "Some of the documents are not exist in the collection"
            )
        async with BulkWriter(session=session) as bulk_writer:
            for document in documents:
                await document.replace(
                    bulk_writer=bulk_writer, session=session
                )

    @wrap_with_actions(EventTypes.UPDATE)
    @save_state_after
    async def update(
        self: Self,
        *args: Union[Dict[Any, Any], Mapping[Any, Any]],
        ignore_revision: bool = False,
        session: Optional[AsyncIOMotorClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        skip_actions: Optional[List[Union[ActionDirections, str]]] = None,
        skip_sync: Optional[bool] = None,
        **pymongo_kwargs: Any,
    ) -> Self:
        """
        Partially update the document in the database

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param session: AsyncIOMotorClientSession - motor session.
        :param ignore_revision: bool - force update. Will update even if revision id is not the same, as stored
        :param bulk_writer: "BulkWriter" - Beanie bulk writer
        :param pymongo_kwargs: pymongo native parameters for update operation
        :return: self
        """
        arguments: list[Any] = list(args)

        if skip_sync is not None:
            raise DeprecationWarning(
                "skip_sync parameter is not supported. The document get synced always using atomic operation."
            )
        use_revision_id = self.get_settings().use_revision

        if self.id is not None:
            find_query: Dict[str, Any] = {"_id": self.id}
        else:
            find_query = {"_id": PydanticObjectId()}

        if use_revision_id and not ignore_revision:
            find_query["revision_id"] = self.revision_id

        if use_revision_id:
            new_revision_id = uuid4()
            arguments.append(SetRevisionId(new_revision_id))
        try:
            result = await self.find_one(find_query).update(
                *arguments,
                session=session,
                response_type=UpdateResponse.NEW_DOCUMENT,
                bulk_writer=bulk_writer,
                **pymongo_kwargs,
            )
        except DuplicateKeyError:
            raise RevisionIdWasChanged
        if bulk_writer is None:
            if use_revision_id and not ignore_revision and result is None:
                raise RevisionIdWasChanged
            merge_models(self, result)
        return self

    @classmethod
    def update_all(
        cls,
        *args: Union[dict, Mapping],
        session: Optional[AsyncIOMotorClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        **pymongo_kwargs: Any,
    ) -> UpdateMany:
        """
        Partially update all the documents

        :param args: *Union[dict, Mapping] - the modifications to apply.
        :param session: AsyncIOMotorClientSession - motor session.
        :param bulk_writer: "BulkWriter" - Beanie bulk writer
        :param **pymongo_kwargs: pymongo native parameters for find operation
        :return: UpdateMany query
        """
        return cls.find_all().update_many(
            *args, session=session, bulk_writer=bulk_writer, **pymongo_kwargs
        )

    def set(
        self: Self,
        expression: Dict[Union[ExpressionField, str, Any], Any],
        session: Optional[AsyncIOMotorClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        skip_sync: Optional[bool] = None,
        **kwargs: Any,
    ) -> Coroutine[None, None, Self]:
        """
        Set values

        Example:

        ```python

        class Sample(Document):
            one: int

        await Document.find(Sample.one == 1).set({Sample.one: 100})

        ```

        Uses [Set operator](operators/update.md#set)

        :param expression: Dict[Union[ExpressionField, str, Any], Any] - keys and
        values to set
        :param session: Optional[AsyncIOMotorClientSession] - motor session
        :param bulk_writer: Optional[BulkWriter] - bulk writer
        :param skip_sync: bool - skip doc syncing. Available for the direct instances only
        :return: self
        """
        return self.update(
            SetOperator(expression),
            session=session,
            bulk_writer=bulk_writer,
            skip_sync=skip_sync,
            **kwargs,
        )

    def current_date(
        self: Self,
        expression: Dict[Union[datetime, ExpressionField, str], Any],
        session: Optional[AsyncIOMotorClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        skip_sync: Optional[bool] = None,
        **kwargs: Any,
    ) -> Coroutine[None, None, Self]:
        """
        Set current date

        Uses [CurrentDate operator](operators/update.md#currentdate)

        :param expression: Dict[Union[datetime, ExpressionField, str], Any]
        :param session: Optional[AsyncIOMotorClientSession] - motor session
        :param bulk_writer: Optional[BulkWriter] - bulk writer
        :param skip_sync: bool - skip doc syncing. Available for the direct instances only
        :return: self
        """
        return self.update(
            CurrentDate(expression),
            session=session,
            bulk_writer=bulk_writer,
            skip_sync=skip_sync,
            **kwargs,
        )

    def inc(
        self: Self,
        expression: Dict[Union[ExpressionField, float, int, str], Any],
        session: Optional[AsyncIOMotorClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        skip_sync: Optional[bool] = None,
        **kwargs: Any,
    ) -> Coroutine[None, None, Self]:
        """
        Increment

        Example:

        ```python

        class Sample(Document):
            one: int

        await Document.find(Sample.one == 1).inc({Sample.one: 100})

        ```

        Uses [Inc operator](operators/update.md#inc)

        :param expression: Dict[Union[ExpressionField, float, int, str], Any]
        :param session: Optional[AsyncIOMotorClientSession] - motor session
        :param bulk_writer: Optional[BulkWriter] - bulk writer
        :param skip_sync: bool - skip doc syncing. Available for the direct instances only
        :return: self
        """
        return self.update(
            Inc(expression),
            session=session,
            bulk_writer=bulk_writer,
            skip_sync=skip_sync,
            **kwargs,
        )

    @wrap_with_actions(EventTypes.DELETE)
    async def delete(
        self,
        session: Optional[AsyncIOMotorClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        link_rule: DeleteRules = DeleteRules.DO_NOTHING,
        skip_actions: Optional[List[Union[ActionDirections, str]]] = None,
        **pymongo_kwargs: Any,
    ) -> Optional[DeleteResult]:
        """
        Delete the document

        :param session: Optional[AsyncIOMotorClientSession] - motor session.
        :param bulk_writer: "BulkWriter" - Beanie bulk writer
        :param link_rule: DeleteRules - rules for link fields
        :param **pymongo_kwargs: pymongo native parameters for delete operation
        :return: Optional[DeleteResult] - pymongo DeleteResult instance.
        """

        if link_rule == DeleteRules.DELETE_LINKS:
            link_fields = self.get_link_fields()
            if link_fields is not None:
                for field_info in link_fields.values():
                    value = getattr(self, field_info.field_name)
                    if field_info.link_type in [
                        LinkTypes.DIRECT,
                        LinkTypes.OPTIONAL_DIRECT,
                        LinkTypes.BACK_DIRECT,
                        LinkTypes.OPTIONAL_BACK_DIRECT,
                    ]:
                        if isinstance(value, Document):
                            await value.delete(
                                link_rule=DeleteRules.DELETE_LINKS,
                                **pymongo_kwargs,
                            )
                    if field_info.link_type in [
                        LinkTypes.LIST,
                        LinkTypes.OPTIONAL_LIST,
                        LinkTypes.BACK_LIST,
                        LinkTypes.OPTIONAL_BACK_LIST,
                    ]:
                        if isinstance(value, List):
                            await asyncio.gather(
                                *[
                                    obj.delete(
                                        link_rule=DeleteRules.DELETE_LINKS,
                                        **pymongo_kwargs,
                                    )
                                    for obj in value
                                    if isinstance(obj, Document)
                                ]
                            )

        return await self.find_one({"_id": self.id}).delete(
            session=session, bulk_writer=bulk_writer, **pymongo_kwargs
        )

    @classmethod
    async def delete_all(
        cls,
        session: Optional[AsyncIOMotorClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        **pymongo_kwargs: Any,
    ) -> Optional[DeleteResult]:
        """
        Delete all the documents

        :param session: Optional[AsyncIOMotorClientSession] - motor session.
        :param bulk_writer: "BulkWriter" - Beanie bulk writer
        :param **pymongo_kwargs: pymongo native parameters for delete operation
        :return: Optional[DeleteResult] - pymongo DeleteResult instance.
        """
        return await cls.find_all().delete(
            session=session, bulk_writer=bulk_writer, **pymongo_kwargs
        )

    # State management

    @classmethod
    def use_state_management(cls) -> bool:
        """
        Is state management turned on
        :return: bool
        """
        return cls.get_settings().use_state_management

    @classmethod
    def state_management_save_previous(cls) -> bool:
        """
        Should we save the previous state after a commit to database
        :return: bool
        """
        return cls.get_settings().state_management_save_previous

    @classmethod
    def state_management_replace_objects(cls) -> bool:
        """
        Should objects be replaced when using state management
        :return: bool
        """
        return cls.get_settings().state_management_replace_objects

    def _save_state(self) -> None:
        """
        Save current document state. Internal method
        :return: None
        """
        if self.use_state_management() and self.id is not None:
            if self.state_management_save_previous():
                self._previous_saved_state = self._saved_state

            self._saved_state = get_dict(
                self,
                to_db=True,
                keep_nulls=self.get_settings().keep_nulls,
                exclude={"revision_id"},
            )

    def get_saved_state(self) -> Optional[Dict[str, Any]]:
        """
        Saved state getter. It is protected property.
        :return: Optional[Dict[str, Any]] - saved state
        """
        return self._saved_state

    def get_previous_saved_state(self) -> Optional[Dict[str, Any]]:
        """
        Previous state getter. It is a protected property.
        :return: Optional[Dict[str, Any]] - previous state
        """
        return self._previous_saved_state

    @property
    @saved_state_needed
    def is_changed(self) -> bool:
        if self._saved_state == get_dict(
            self,
            to_db=True,
            keep_nulls=self.get_settings().keep_nulls,
            exclude={"revision_id"},
        ):
            return False
        return True

    @property
    @saved_state_needed
    @previous_saved_state_needed
    def has_changed(self) -> bool:
        if (
            self._previous_saved_state is None
            or self._previous_saved_state == self._saved_state
        ):
            return False
        return True

    def _collect_updates(
        self, old_dict: Dict[str, Any], new_dict: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Compares old_dict with new_dict and returns field paths that have been updated
        Args:
            old_dict: dict1
            new_dict: dict2

        Returns: dictionary with updates

        """
        updates = {}
        if old_dict.keys() - new_dict.keys():
            updates = new_dict
        else:
            for field_name, field_value in new_dict.items():
                if field_value != old_dict.get(field_name):
                    if not self.state_management_replace_objects() and (
                        isinstance(field_value, dict)
                        and isinstance(old_dict.get(field_name), dict)
                    ):
                        if old_dict.get(field_name) is None:
                            updates[field_name] = field_value
                        elif isinstance(field_value, dict) and isinstance(
                            old_dict.get(field_name), dict
                        ):
                            field_data = self._collect_updates(
                                old_dict.get(field_name),  # type: ignore
                                field_value,
                            )

                            for k, v in field_data.items():
                                updates[f"{field_name}.{k}"] = v
                    else:
                        updates[field_name] = field_value

        return updates

    @saved_state_needed
    def get_changes(self) -> Dict[str, Any]:
        return self._collect_updates(
            self._saved_state,  # type: ignore
            get_dict(
                self,
                to_db=True,
                keep_nulls=self.get_settings().keep_nulls,
                exclude={"revision_id"},
            ),
        )

    @saved_state_needed
    @previous_saved_state_needed
    def get_previous_changes(self) -> Dict[str, Any]:
        if self._previous_saved_state is None:
            return {}

        return self._collect_updates(
            self._previous_saved_state,
            self._saved_state,  # type: ignore
        )

    @saved_state_needed
    def rollback(self) -> None:
        if self.is_changed:
            for key, value in self._saved_state.items():  # type: ignore
                if key == "_id":
                    setattr(self, "id", value)
                else:
                    setattr(self, key, value)

    # Other

    @classmethod
    def get_settings(cls) -> DocumentSettings:
        """
        Get document settings, which was created on
        the initialization step

        :return: DocumentSettings class
        """
        if cls._document_settings is None:
            raise CollectionWasNotInitialized
        return cls._document_settings

    @classmethod
    async def inspect_collection(
        cls, session: Optional[AsyncIOMotorClientSession] = None
    ) -> InspectionResult:
        """
        Check, if documents, stored in the MongoDB collection
        are compatible with the Document schema

        :return: InspectionResult
        """
        inspection_result = InspectionResult()
        async for json_document in cls.get_motor_collection().find(
            {}, session=session
        ):
            try:
                parse_model(cls, json_document)
            except ValidationError as e:
                if inspection_result.status == InspectionStatuses.OK:
                    inspection_result.status = InspectionStatuses.FAIL
                inspection_result.errors.append(
                    InspectionError(
                        document_id=json_document["_id"], error=str(e)
                    )
                )
        return inspection_result

    @classmethod
    def _check_hidden_fields(cls):
        hidden_fields = [
            (name, field)
            for name, field in get_model_fields(cls).items()
            if get_extra_field_info(field, "hidden") is True
        ]
        if not hidden_fields:
            return
        warnings.warn(
            f"{cls.__name__}: 'hidden=True' is deprecated, please use 'exclude=True'",
            DeprecationWarning,
            stacklevel=2,
        )
        if IS_PYDANTIC_V2:
            for name, field in hidden_fields:
                field.exclude = True
                del field.json_schema_extra["hidden"]
            cls.model_rebuild(force=True)
        else:
            for name, field in hidden_fields:
                field.field_info.exclude = True
                del field.field_info.extra["hidden"]
                cls.__exclude_fields__[name] = True

    @wrap_with_actions(event_type=EventTypes.VALIDATE_ON_SAVE)
    async def validate_self(self, *args: Any, **kwargs: Any):
        # TODO: it can be sync, but needs some actions controller improvements
        if self.get_settings().validate_on_save:
            new_model = parse_model(self.__class__, get_model_dump(self))
            merge_models(self, new_model)

    def to_ref(self):
        if self.id is None:
            raise DocumentWasNotSaved("Can not create dbref without id")
        return DBRef(self.get_motor_collection().name, self.id)

    async def fetch_link(self, field: Union[str, Any]):
        ref_obj = getattr(self, field, None)
        if isinstance(ref_obj, Link):
            value = await ref_obj.fetch(fetch_links=True)
            setattr(self, field, value)
        if isinstance(ref_obj, list) and ref_obj:
            values = await Link.fetch_list(ref_obj, fetch_links=True)
            setattr(self, field, values)

    async def fetch_all_links(self):
        coros = []
        link_fields = self.get_link_fields()
        if link_fields is not None:
            for ref in link_fields.values():
                coros.append(self.fetch_link(ref.field_name))  # TODO lists
        await asyncio.gather(*coros)

    @classmethod
    def get_link_fields(cls) -> Optional[Dict[str, LinkInfo]]:
        return cls._link_fields

    @classmethod
    def get_model_type(cls) -> ModelType:
        return ModelType.Document

    @classmethod
    async def distinct(
        cls,
        key: str,
        filter: Optional[Mapping[str, Any]] = None,
        session: Optional[AsyncIOMotorClientSession] = None,
        **kwargs: Any,
    ) -> list:
        return await cls.get_motor_collection().distinct(
            key, filter, session, **kwargs
        )

    @classmethod
    def link_from_id(cls, id: Any):
        ref = DBRef(id=id, collection=cls.get_collection_name())
        return Link(ref, document_class=cls)

    @classmethod
    def bulk_writer(
        cls,
        session: Optional[AsyncIOMotorClientSession] = None,
        ordered: bool = True,
        bypass_document_validation: bool = False,
        comment: Optional[Any] = None,
    ) -> BulkWriter:
        """
        Returns a BulkWriter instance for handling bulk write operations.

        :param session: ClientSession
            The session instance used for transactional operations.
        :param ordered: bool
            If ``True`` (the default), requests will be performed on the server serially, in the order provided. If an error
            occurs, all remaining operations are aborted. If ``False``, requests will be performed on the server in
            arbitrary order, possibly in parallel, and all operations will be attempted.
        :param bypass_document_validation: bool, optional
            If ``True``, allows the write to opt-out of document-level validation. Default is ``False``.
        :param comment: str, optional
            A user-provided comment to attach to the BulkWriter.

        :returns: BulkWriter
            An instance of BulkWriter configured with the provided settings.

        Example Usage:
        --------------
        This method is typically used within an asynchronous context manager.

        .. code-block:: python

            async with Document.bulk_writer(ordered=True) as bulk:
                await Document.insert_one(Document(field="value"), bulk_writer=bulk)
        """
        return BulkWriter(
            session, ordered, cls, bypass_document_validation, comment
        )


class DocumentWithSoftDelete(Document):
    deleted_at: Optional[datetime] = None

    def is_deleted(self) -> bool:
        return self.deleted_at is not None

    async def hard_delete(
        self,
        session: Optional[AsyncIOMotorClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        link_rule: DeleteRules = DeleteRules.DO_NOTHING,
        skip_actions: Optional[List[Union[ActionDirections, str]]] = None,
        **pymongo_kwargs: Any,
    ) -> Optional[DeleteResult]:
        return await super().delete(
            session=session,
            bulk_writer=bulk_writer,
            link_rule=link_rule,
            skip_actions=skip_actions,
            **pymongo_kwargs,
        )

    async def delete(
        self,
        session: Optional[AsyncIOMotorClientSession] = None,
        bulk_writer: Optional[BulkWriter] = None,
        link_rule: DeleteRules = DeleteRules.DO_NOTHING,
        skip_actions: Optional[List[Union[ActionDirections, str]]] = None,
        **pymongo_kwargs,
    ) -> Optional[DeleteResult]:
        self.deleted_at = datetime.now(tz=timezone.utc)
        await self.save()
        return None

    @classmethod
    def find_many_in_all(  # type: ignore
        cls: Type[FindType],
        *args: Union[Mapping[str, Any], bool],
        projection_model: Optional[Type["DocumentProjectionType"]] = None,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        sort: Union[None, str, List[Tuple[str, SortDirection]]] = None,
        session: Optional[AsyncIOMotorClientSession] = None,
        ignore_cache: bool = False,
        fetch_links: bool = False,
        with_children: bool = False,
        lazy_parse: bool = False,
        nesting_depth: Optional[int] = None,
        nesting_depths_per_field: Optional[Dict[str, int]] = None,
        **pymongo_kwargs: Any,
    ) -> Union[FindMany[FindType], FindMany["DocumentProjectionType"]]:
        return cls._find_many_query_class(document_model=cls).find_many(
            *args,
            sort=sort,
            skip=skip,
            limit=limit,
            projection_model=projection_model,
            session=session,
            ignore_cache=ignore_cache,
            fetch_links=fetch_links,
            lazy_parse=lazy_parse,
            nesting_depth=nesting_depth,
            nesting_depths_per_field=nesting_depths_per_field,
            **pymongo_kwargs,
        )

    @classmethod
    def find_many(  # type: ignore
        cls: Type[FindType],
        *args: Union[Mapping[str, Any], bool],
        projection_model: Optional[Type["DocumentProjectionType"]] = None,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        sort: Union[None, str, List[Tuple[str, SortDirection]]] = None,
        session: Optional[AsyncIOMotorClientSession] = None,
        ignore_cache: bool = False,
        fetch_links: bool = False,
        with_children: bool = False,
        lazy_parse: bool = False,
        nesting_depth: Optional[int] = None,
        nesting_depths_per_field: Optional[Dict[str, int]] = None,
        **pymongo_kwargs: Any,
    ) -> Union[FindMany[FindType], FindMany["DocumentProjectionType"]]:
        args = cls._add_class_id_filter(args, with_children) + (
            {"deleted_at": None},
        )
        return cls._find_many_query_class(document_model=cls).find_many(
            *args,
            sort=sort,
            skip=skip,
            limit=limit,
            projection_model=projection_model,
            session=session,
            ignore_cache=ignore_cache,
            fetch_links=fetch_links,
            lazy_parse=lazy_parse,
            nesting_depth=nesting_depth,
            nesting_depths_per_field=nesting_depths_per_field,
            **pymongo_kwargs,
        )

    @classmethod
    def find_one(  # type: ignore
        cls: Type[FindType],
        *args: Union[Mapping[str, Any], bool],
        projection_model: Optional[Type["DocumentProjectionType"]] = None,
        session: Optional[AsyncIOMotorClientSession] = None,
        ignore_cache: bool = False,
        fetch_links: bool = False,
        with_children: bool = False,
        nesting_depth: Optional[int] = None,
        nesting_depths_per_field: Optional[Dict[str, int]] = None,
        **pymongo_kwargs: Any,
    ) -> Union[FindOne[FindType], FindOne["DocumentProjectionType"]]:
        args = cls._add_class_id_filter(args, with_children) + (
            {"deleted_at": None},
        )
        return cls._find_one_query_class(document_model=cls).find_one(
            *args,
            projection_model=projection_model,
            session=session,
            ignore_cache=ignore_cache,
            fetch_links=fetch_links,
            nesting_depth=nesting_depth,
            nesting_depths_per_field=nesting_depths_per_field,
            **pymongo_kwargs,
        )



from typing import List

import pytest
from pydantic.fields import Field

from beanie import Document, init_beanie
from beanie.exceptions import DocumentWasNotSaved
from beanie.odm.fields import (
    BackLink,
    DeleteRules,
    Link,
    WriteRules,
)
from beanie.odm.utils.pydantic import (
    IS_PYDANTIC_V2,
    get_model_fields,
    parse_model,
)
from beanie.operators import In, Or
from tests.odm.models import (
    AddressView,
    ADocument,
    BDocument,
    DocumentToBeLinked,
    DocumentWithBackLink,
    DocumentWithBackLinkForNesting,
    DocumentWithLink,
    DocumentWithLinkForNesting,
    DocumentWithListBackLink,
    DocumentWithListLink,
    DocumentWithListOfLinks,
    DocumentWithTextIndexAndLink,
    Door,
    House,
    LinkDocumentForTextSeacrh,
    Lock,
    LongSelfLink,
    LoopedLinksA,
    LoopedLinksB,
    Region,
    Roof,
    RootDocument,
    SelfLinked,
    UsersAddresses,
    Window,
    Yard,
)


def lock_not_inserted_fn():
    return Lock(k=10)


@pytest.fixture
def locks_not_inserted():
    return [Lock(k=10001), Lock(k=20002)]


@pytest.fixture
def window_not_inserted():
    return Window(x=10, y=10, lock=lock_not_inserted_fn())


@pytest.fixture
def windows_not_inserted():
    return [
        Window(
            x=10,
            y=10,
            lock=lock_not_inserted_fn(),
        ),
        Window(
            x=11,
            y=11,
            lock=lock_not_inserted_fn(),
        ),
    ]


@pytest.fixture
def door_not_inserted(window_not_inserted, locks_not_inserted):
    return Door(t=10, window=window_not_inserted, locks=locks_not_inserted)


@pytest.fixture
def house_not_inserted(windows_not_inserted, door_not_inserted):
    return House(
        windows=windows_not_inserted, door=door_not_inserted, name="test"
    )


@pytest.fixture
async def house(house_not_inserted):
    return await house_not_inserted.insert(link_rule=WriteRules.WRITE)


@pytest.fixture
async def houses():
    for i in range(10):
        roof = Roof() if i % 2 == 0 else None
        if i % 2 == 0:
            yards = [Yard(v=10, w=10 + i), Yard(v=11, w=10 + i)]
        else:
            yards = None
        house = await House(
            door=Door(
                t=i,
                window=Window(x=20, y=21 + i, lock=Lock(k=20 + i))
                if i % 2 == 0
                else None,
                locks=[Lock(k=20 + i)],
            ),
            windows=[
                Window(x=10, y=10 + i, lock=Lock(k=10 + i)),
                Window(x=11, y=11 + i, lock=Lock(k=11 + i)),
            ],
            yards=yards,
            roof=roof,
            name="test",
            height=i,
        ).insert(link_rule=WriteRules.WRITE)
        if i == 9:
            await house.windows[0].delete()
            await house.windows[1].lock.delete()
            await house.door.delete()


class TestInsert:
    async def test_rule_do_nothing(self, house_not_inserted):
        with pytest.raises(DocumentWasNotSaved):
            await house_not_inserted.insert()

    async def test_rule_write(self, house_not_inserted):
        await house_not_inserted.insert(link_rule=WriteRules.WRITE)
        locks = await Lock.all().to_list()
        assert len(locks) == 5
        windows = await Window.all().to_list()
        assert len(windows) == 3
        doors = await Door.all().to_list()
        assert len(doors) == 1
        houses = await House.all().to_list()
        assert len(houses) == 1

    async def test_insert_with_link(
        self,
        house_not_inserted,
        door_not_inserted,
        window_not_inserted,
        locks_not_inserted,
    ):
        lock_links = []
        for lock in locks_not_inserted:
            lock = await lock.insert()
            link = Lock.link_from_id(lock.id)
            lock_links.append(link)
        door_not_inserted.locks = lock_links

        door_window_lock = await lock_not_inserted_fn().insert()
        door_window_lock_link = Lock.link_from_id(door_window_lock.id)
        window_not_inserted.lock = door_window_lock_link

        door_window = await window_not_inserted.insert()
        door_window_link = Window.link_from_id(door_window.id)
        door_not_inserted.window = door_window_link

        door = await door_not_inserted.insert()
        door_link = Door.link_from_id(door.id)
        house_not_inserted.door = door_link

        house = parse_model(House, house_not_inserted)
        await house.insert(link_rule=WriteRules.WRITE)

        if IS_PYDANTIC_V2:
            json_str = house.model_dump_json()
        else:
            json_str = house.json()
        assert json_str is not None

    async def test_multi_insert_links(self):
        house = House(name="random", windows=[], door=Door())
        window = await Window(x=13, y=23).insert()
        assert window.id
        house.windows.append(window)

        house = await house.insert(link_rule=WriteRules.WRITE)
        new_window_1 = Window(x=11, y=22)
        assert new_window_1.id is None
        house.windows.append(new_window_1)
        new_window_2 = Window(x=12, y=23)
        assert new_window_2.id is None
        house.windows.append(new_window_2)
        await house.save(link_rule=WriteRules.WRITE)
        for win in house.windows:
            assert isinstance(win, Window)
            assert win.id
        assert new_window_1.id is not None
        assert new_window_2.id is not None

    async def test_fetch_after_insert(self, house_not_inserted):
        # TODO: what is the point of this test if nothing was inserted to DB?
        await house_not_inserted.fetch_all_links()


class TestFind:
    async def test_prefetch_find_many(self, houses):
        items = await House.find(House.height > 2).sort(House.height).to_list()
        assert len(items) == 7
        for window in items[0].windows:
            assert isinstance(window, Link)
        assert items[0].yards is None
        for yard in items[1].yards:
            assert isinstance(yard, Link)
        assert isinstance(items[0].door, Link)
        assert items[0].roof is None
        assert isinstance(items[1].roof, Link)

        items = (
            await House.find(House.height > 2, fetch_links=True)
            .sort(House.height)
            .to_list()
        )
        assert len(items) == 7
        for window in items[0].windows:
            assert isinstance(window, Window)
            assert isinstance(window.lock, Lock)
        assert items[0].yards == []
        for yard in items[1].yards:
            assert isinstance(yard, Yard)
        assert isinstance(items[0].door, Door)
        assert isinstance(items[1].door.window, Window)
        assert items[0].door.window is None
        assert isinstance(items[1].door.window.lock, Lock)
        for lock in items[0].door.locks:
            assert isinstance(lock, Lock)
        assert items[0].roof is None
        assert isinstance(items[1].roof, Roof)

        houses = await House.find_many(
            House.height == 9, fetch_links=True
        ).to_list()
        assert len(houses[0].windows) == 1
        assert isinstance(houses[0].windows[0].lock, Link)
        assert isinstance(houses[0].door, Link)

        await houses[0].fetch_link(House.door)
        assert isinstance(houses[0].door, Link)

        houses = await House.find_many(
            House.door.t > 5, fetch_links=True
        ).to_list()

        assert len(houses) == 3

        houses = await House.find_many(
            House.windows.y == 15, fetch_links=True
        ).to_list()

        assert len(houses) == 2

        houses = await House.find_many(
            House.height > 5, limit=3, fetch_links=True
        ).to_list()

        assert len(houses) == 3

    async def test_prefect_count(self, houses):
        c = await House.find(House.door.t > 5, fetch_links=True).count()
        assert c == 3

        c = await House.find_one(House.door.t > 5, fetch_links=True).count()
        assert c == 3

    async def test_prefetch_find_one(self, house):
        house = await House.find_one(House.name == "test")
        for window in house.windows:
            assert isinstance(window, Link)
        assert isinstance(house.door, Link)

        house = await House.find_one(House.name == "test", fetch_links=True)
        for window in house.windows:
            assert isinstance(window, Window)
        assert isinstance(house.door, Door)

        house = await House.get(house.id, fetch_links=True)
        for window in house.windows:
            assert isinstance(window, Window)
        assert isinstance(house.door, Door)

    async def test_fetch(self, house):
        house = await House.find_one(House.name == "test")
        for window in house.windows:
            assert isinstance(window, Link)
        assert isinstance(house.door, Link)

        await house.fetch_all_links()
        for window in house.windows:
            assert isinstance(window, Window)
            assert isinstance(window.lock, Lock)
        assert isinstance(house.door, Door)
        assert isinstance(house.door.window, Window)
        for lock in house.door.locks:
            assert isinstance(lock, Lock)

        house = await House.find_one(House.name == "test")
        assert isinstance(house.door, Link)
        await house.fetch_link(House.door)
        assert isinstance(house.door, Door)
        assert isinstance(house.door.window, Window)
        for lock in house.door.locks:
            assert isinstance(lock, Lock)

        for window in house.windows:
            assert isinstance(window, Link)
        await house.fetch_link(House.windows)
        for window in house.windows:
            assert isinstance(window, Window)
            assert isinstance(window.lock, Lock)

    async def test_find_by_id_of_the_linked_docs(self, house):
        house_lst_1 = await House.find(
            House.door.id == house.door.id
        ).to_list()
        house_lst_2 = await House.find(
            House.door.id == house.door.id, fetch_links=True
        ).to_list()
        assert len(house_lst_1) == 1
        assert len(house_lst_2) == 1

        house_1 = await House.find_one(House.door.id == house.door.id)
        house_2 = await House.find_one(
            House.door.id == house.door.id, fetch_links=True
        )
        assert house_1 is not None
        assert house_2 is not None

    async def test_find_by_id_list_of_the_linked_docs(self, houses):
        items = (
            await House.find(House.height < 3, fetch_links=True)
            .sort(House.height)
            .to_list()
        )
        assert len(items) == 3

        house_lst_1 = await House.find(
            Or(
                House.door.id == items[0].door.id,
                In(House.door.id, [items[1].door.id, items[2].door.id]),
            )
        ).to_list()
        house_lst_2 = await House.find(
            Or(
                House.door.id == items[0].door.id,
                In(House.door.id, [items[1].door.id, items[2].door.id]),
            ),
            fetch_links=True,
        ).to_list()

        assert len(house_lst_1) == 3
        assert len(house_lst_2) == 3

    async def test_fetch_list_with_some_prefetched(self):
        docs = []
        for i in range(10):
            doc = DocumentToBeLinked()
            await doc.save()
            docs.append(doc)

        doc_with_links = DocumentWithListOfLinks(links=docs)
        await doc_with_links.save()

        doc_with_links = await DocumentWithListOfLinks.get(
            doc_with_links.id, fetch_links=False
        )
        doc_with_links.links[-1] = await doc_with_links.links[-1].fetch()

        await doc_with_links.fetch_all_links()

        for link in doc_with_links.links:
            assert isinstance(link, DocumentToBeLinked)

        assert len(doc_with_links.links) == 10

        # test order
        for i in range(10):
            assert doc_with_links.links[i].id == docs[i].id

    async def test_text_search(self):
        doc = DocumentWithTextIndexAndLink(
            s="hello world", link=LinkDocumentForTextSeacrh(i=1)
        )
        await doc.insert(link_rule=WriteRules.WRITE)

        doc2 = DocumentWithTextIndexAndLink(
            s="hi world", link=LinkDocumentForTextSeacrh(i=2)
        )
        await doc2.insert(link_rule=WriteRules.WRITE)

        docs = await DocumentWithTextIndexAndLink.find(
            {"$text": {"$search": "hello"}}, fetch_links=True
        ).to_list()
        assert len(docs) == 1

    async def test_self_nesting_find_parameters(self):
        self_linked_doc = LongSelfLink()
        await self_linked_doc.insert(link_rule=WriteRules.WRITE)
        self_linked_doc.link = self_linked_doc
        await self_linked_doc.save()

        self_linked_doc = await LongSelfLink.find_one(
            nesting_depth=4, fetch_links=True
        )
        assert self_linked_doc.link.link.link.link.id == self_linked_doc.id
        assert isinstance(self_linked_doc.link.link.link.link.link, Link)

        self_linked_doc = await LongSelfLink.find_one(
            nesting_depth=0, fetch_links=True
        )
        assert isinstance(self_linked_doc.link, Link)

    async def test_nesting_find_parameters(self):
        back_link_doc = DocumentWithBackLinkForNesting(i=1)
        await back_link_doc.insert()
        link_doc = DocumentWithLinkForNesting(link=back_link_doc, s="TEST")
        await link_doc.insert()

        doc = await DocumentWithBackLinkForNesting.find_one(
            DocumentWithBackLinkForNesting.i == 1,
            fetch_links=True,
            nesting_depths_per_field={"back_link": 2},
        )
        assert doc.back_link.link.id == doc.id
        assert isinstance(doc.back_link.link.back_link, BackLink)


class TestReplace:
    async def test_do_nothing(self, house):
        house.door.t = 100
        await house.replace()
        new_house = await House.get(house.id, fetch_links=True)
        assert new_house.door.t == 10

    async def test_write(self, house):
        house.door.t = 100
        await house.replace(link_rule=WriteRules.WRITE)
        new_house = await House.get(house.id, fetch_links=True)
        assert new_house.door.t == 100


class TestSave:
    async def test_do_nothing(self, house):
        house.door.t = 100
        await house.save()
        new_house = await House.get(house.id, fetch_links=True)
        assert new_house.door.t == 10

    async def test_write(self, house):
        house.door.t = 100
        new_window = Window(x=100, y=100, lock=Lock(k=100))
        house.windows = [new_window]
        assert new_window.id is None
        await house.save(link_rule=WriteRules.WRITE)
        new_house = await House.get(house.id, fetch_links=True)
        assert new_house.door.t == 100
        for window in new_house.windows:
            assert window.x == 100
            assert window.y == 100
            assert window.id is not None
            assert isinstance(window.lock, Lock)
            assert window.lock.k == 100
            assert window.lock.id is not None
        assert new_window.id is not None


class TestDelete:
    async def test_do_nothing(self, house):
        await house.delete()
        door = await Door.get(house.door.id)
        assert door is not None

        windows = await Window.all().to_list()
        assert windows is not None

        locks = await Lock.all().to_list()
        assert locks is not None

    async def test_delete_links(self, house):
        await house.delete(link_rule=DeleteRules.DELETE_LINKS)
        door = await Door.get(house.door.id)
        assert door is None

        windows = await Window.all().to_list()
        assert windows == []

        locks = await Lock.all().to_list()
        assert locks == []


class TestOther:
    async def test_query_composition(self):
        SYS = {"id", "revision_id"}

        # Simple fields are initialized using the pydantic model_fields internal property
        # such fields are properly isolated when multi inheritance is involved.
        assert set(get_model_fields(RootDocument).keys()) == SYS | {
            "name",
            "link_root",
        }
        assert set(get_model_fields(ADocument).keys()) == SYS | {
            "name",
            "link_root",
            "surname",
            "link_a",
        }
        assert set(get_model_fields(BDocument).keys()) == SYS | {
            "name",
            "link_root",
            "email",
            "link_b",
        }

        # Where Document.init_fields() has a bug that prevents proper link inheritance when parent
        # documents are initialized. Furthermore, some-why BDocument._link_fields are not deterministic
        assert set(RootDocument._link_fields.keys()) == {"link_root"}
        assert set(ADocument._link_fields.keys()) == {"link_root", "link_a"}
        assert set(BDocument._link_fields.keys()) == {"link_root", "link_b"}

    async def test_with_projection(self):
        await UsersAddresses(region_id=Region()).insert(
            link_rule=WriteRules.WRITE
        )
        res = await UsersAddresses.find_one(fetch_links=True).project(
            AddressView
        )
        assert res.id is not None
        assert res.state == "TEST"
        assert res.city == "TEST"

    async def test_self_linked(self):
        await SelfLinked(item=SelfLinked(s="2"), s="1").insert(
            link_rule=WriteRules.WRITE
        )

        res = await SelfLinked.find_one(fetch_links=True)
        assert isinstance(res, SelfLinked)
        assert res.item is None

        await SelfLinked.delete_all()

        await SelfLinked(
            item=SelfLinked(
                item=SelfLinked(item=SelfLinked(s="4"), s="3"), s="2"
            ),
            s="1",
        ).insert(link_rule=WriteRules.WRITE)

        res = await SelfLinked.find_one(SelfLinked.s == "1", fetch_links=True)
        assert isinstance(res, SelfLinked)
        assert isinstance(res.item, SelfLinked)
        assert isinstance(res.item.item, SelfLinked)
        assert isinstance(res.item.item.item, Link)

    async def test_looped_links(self):
        await LoopedLinksA(
            b=LoopedLinksB(
                a=LoopedLinksA(
                    b=LoopedLinksB(
                        s="4",
                    ),
                    s="3",
                ),
                s="2",
            ),
            s="1",
        ).insert(link_rule=WriteRules.WRITE)
        res = await LoopedLinksA.find_one(
            LoopedLinksA.s == "1", fetch_links=True
        )
        assert isinstance(res, LoopedLinksA)
        assert isinstance(res.b, LoopedLinksB)
        assert isinstance(res.b.a, LoopedLinksA)
        assert isinstance(res.b.a.b, Link)

        await LoopedLinksA(
            b=LoopedLinksB(s="a2"),
            s="a1",
        ).insert(link_rule=WriteRules.WRITE)
        res = await LoopedLinksA.find_one(
            LoopedLinksA.s == "a1", fetch_links=True
        )
        assert isinstance(res, LoopedLinksA)
        assert isinstance(res.b, LoopedLinksB)
        assert res.b.a is None

    async def test_with_chaining_aggregation(self):
        region = Region()
        await region.insert()

        for i in range(10):
            await UsersAddresses(region_id=region).insert()

        region_2 = Region()
        await region_2.insert()

        for i in range(10):
            await UsersAddresses(region_id=region_2).insert()

        addresses_count = (
            await UsersAddresses.find(
                UsersAddresses.region_id.id == region.id, fetch_links=True
            )
            .aggregate([{"$count": "count"}])
            .to_list()
        )

        assert addresses_count[0] == {"count": 10}

    async def test_with_chaining_aggregation_and_text_search(self):
        # ARRANGE
        NUM_DOCS = 10
        NUM_WITH_LOWER = 5
        linked_document = LinkDocumentForTextSeacrh(i=1)
        await linked_document.insert()

        for i in range(NUM_DOCS):
            await DocumentWithTextIndexAndLink(
                s="lower" if i < NUM_WITH_LOWER else "UPPER",
                link=linked_document,
            ).insert()

        linked_document_2 = LinkDocumentForTextSeacrh(i=2)
        await linked_document_2.insert()

        for i in range(NUM_DOCS):
            await DocumentWithTextIndexAndLink(
                s="lower" if i < NUM_WITH_LOWER else "UPPER",
                link=linked_document_2,
            ).insert()

        # ACT
        query = DocumentWithTextIndexAndLink.find(
            {"$text": {"$search": "lower"}},
            DocumentWithTextIndexAndLink.link.i == 1,
            fetch_links=True,
        )

        # Test both aggregation and count methods
        document_count_aggregation = await query.aggregate(
            [{"$count": "count"}]
        ).to_list()
        document_count = await query.count()

        # ASSERT
        assert document_count_aggregation[0] == {"count": NUM_WITH_LOWER}
        assert document_count == NUM_WITH_LOWER

    async def test_with_extra_allow(self, houses):
        res = await House.find(fetch_links=True).to_list()
        assert get_model_fields(res[0]).keys() == {
            "id",
            "revision_id",
            "windows",
            "door",
            "roof",
            "yards",
            "name",
            "height",
        }

        res = await House.find_one(fetch_links=True)
        assert get_model_fields(res).keys() == {
            "id",
            "revision_id",
            "windows",
            "door",
            "roof",
            "yards",
            "name",
            "height",
        }


@pytest.fixture()
async def link_and_backlink_doc_pair():
    back_link_doc = DocumentWithBackLink()
    await back_link_doc.insert()
    link_doc = DocumentWithLink(link=back_link_doc)
    await link_doc.insert()
    return link_doc, back_link_doc


@pytest.fixture()
async def list_link_and_list_backlink_doc_pair():
    back_link_doc = DocumentWithListBackLink()
    await back_link_doc.insert()
    link_doc = DocumentWithListLink(link=[back_link_doc])
    await link_doc.insert()
    return link_doc, back_link_doc


class TestFindBackLinks:
    async def test_prefetch_direct(self, link_and_backlink_doc_pair):
        link_doc, back_link_doc = link_and_backlink_doc_pair
        back_link_doc = await DocumentWithBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        assert back_link_doc.back_link.id == link_doc.id
        assert back_link_doc.back_link.link.id == back_link_doc.id

    async def test_prefetch_list(self, list_link_and_list_backlink_doc_pair):
        link_doc, back_link_doc = list_link_and_list_backlink_doc_pair
        back_link_doc = await DocumentWithListBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        assert back_link_doc.back_link[0].id == link_doc.id
        assert back_link_doc.back_link[0].link[0].id == back_link_doc.id

    async def test_nesting(self):
        back_link_doc = DocumentWithBackLinkForNesting(i=1)
        await back_link_doc.insert()
        link_doc = DocumentWithLinkForNesting(link=back_link_doc, s="TEST")
        await link_doc.insert()

        doc = await DocumentWithLinkForNesting.get(
            link_doc.id, fetch_links=True
        )
        assert isinstance(doc.link, Link)
        doc.link = await doc.link.fetch()
        assert doc.link.i == 1

        back_link_doc = await DocumentWithBackLinkForNesting.get(
            back_link_doc.id, fetch_links=True
        )
        assert (
            back_link_doc.back_link.link.back_link.link.back_link.id
            == link_doc.id
        )
        assert isinstance(
            back_link_doc.back_link.link.back_link.link.back_link.link, Link
        )


class TestReplaceBackLinks:
    async def test_do_nothing(self, link_and_backlink_doc_pair):
        link_doc, back_link_doc = link_and_backlink_doc_pair
        back_link_doc.back_link.s = "new value"
        await back_link_doc.replace()
        new_back_link_doc = await DocumentWithBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        assert new_back_link_doc.back_link.s == "TEST"

    async def test_do_nothing_list(self, list_link_and_list_backlink_doc_pair):
        link_doc, back_link_doc = list_link_and_list_backlink_doc_pair
        back_link_doc = await DocumentWithListBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        for lnk in back_link_doc.back_link:
            lnk.s = "new value"
        await back_link_doc.replace()
        new_back_link_doc = await DocumentWithListBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        for lnk in new_back_link_doc.back_link:
            assert lnk.s == "TEST"

    async def test_write(self, link_and_backlink_doc_pair):
        link_doc, back_link_doc = link_and_backlink_doc_pair
        back_link_doc = await DocumentWithBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        back_link_doc.back_link.s = "new value"
        await back_link_doc.replace(link_rule=WriteRules.WRITE)
        new_back_link_doc = await DocumentWithBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        assert new_back_link_doc.back_link.s == "new value"

    async def test_do_nothing_write_list(
        self, list_link_and_list_backlink_doc_pair
    ):
        link_doc, back_link_doc = list_link_and_list_backlink_doc_pair
        back_link_doc = await DocumentWithListBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        for lnk in back_link_doc.back_link:
            lnk.s = "new value"
        await back_link_doc.replace(link_rule=WriteRules.WRITE)
        new_back_link_doc = await DocumentWithListBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        for lnk in new_back_link_doc.back_link:
            assert lnk.s == "new value"


class TestSaveBackLinks:
    async def test_do_nothing(self, link_and_backlink_doc_pair):
        link_doc, back_link_doc = link_and_backlink_doc_pair
        back_link_doc.back_link.s = "new value"
        await back_link_doc.save()
        new_back_link_doc = await DocumentWithBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        assert new_back_link_doc.back_link.s == "TEST"

    async def test_do_nothing_list(self, list_link_and_list_backlink_doc_pair):
        link_doc, back_link_doc = list_link_and_list_backlink_doc_pair
        back_link_doc = await DocumentWithListBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        for lnk in back_link_doc.back_link:
            lnk.s = "new value"
        await back_link_doc.save()
        new_back_link_doc = await DocumentWithListBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        for lnk in new_back_link_doc.back_link:
            assert lnk.s == "TEST"

    async def test_write(self, link_and_backlink_doc_pair):
        link_doc, back_link_doc = link_and_backlink_doc_pair
        back_link_doc = await DocumentWithBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        back_link_doc.back_link.s = "new value"
        await back_link_doc.save(link_rule=WriteRules.WRITE)
        new_back_link_doc = await DocumentWithBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        assert new_back_link_doc.back_link.s == "new value"

    async def test_write_list(self, list_link_and_list_backlink_doc_pair):
        link_doc, back_link_doc = list_link_and_list_backlink_doc_pair
        back_link_doc = await DocumentWithListBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        for lnk in back_link_doc.back_link:
            lnk.s = "new value"
        await back_link_doc.save(link_rule=WriteRules.WRITE)
        new_back_link_doc = await DocumentWithListBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        for lnk in new_back_link_doc.back_link:
            assert lnk.s == "new value"


class HouseForReversedOrderInit(Document):
    name: str
    door: Link["DoorForReversedOrderInit"]
    owners: List[Link["PersonForReversedOrderInit"]]


class DoorForReversedOrderInit(Document):
    height: int = 2
    width: int = 1
    if IS_PYDANTIC_V2:
        house: BackLink[HouseForReversedOrderInit] = Field(
            json_schema_extra={"original_field": "door"}
        )
    else:
        house: BackLink[HouseForReversedOrderInit] = Field(
            original_field="door"
        )


class PersonForReversedOrderInit(Document):
    name: str
    if IS_PYDANTIC_V2:
        house: List[BackLink[HouseForReversedOrderInit]] = Field(
            json_schema_extra={"original_field": "owners"}
        )
    else:
        house: List[BackLink[HouseForReversedOrderInit]] = Field(
            original_field="owners"
        )


class TestDeleteBackLinks:
    async def test_do_nothing(self, link_and_backlink_doc_pair):
        link_doc, back_link_doc = link_and_backlink_doc_pair
        back_link_doc = await DocumentWithBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        await back_link_doc.delete()
        new_link_doc = await DocumentWithLink.get(
            link_doc.id, fetch_links=True
        )
        assert new_link_doc is not None

    async def test_do_nothing_list(self, list_link_and_list_backlink_doc_pair):
        link_doc, back_link_doc = list_link_and_list_backlink_doc_pair
        back_link_doc = await DocumentWithListBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        await back_link_doc.delete()
        new_link_doc = await DocumentWithListLink.get(
            link_doc.id, fetch_links=True
        )
        assert new_link_doc is not None

    async def test_delete_links(self, link_and_backlink_doc_pair):
        link_doc, back_link_doc = link_and_backlink_doc_pair
        back_link_doc = await DocumentWithBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        await back_link_doc.delete(link_rule=DeleteRules.DELETE_LINKS)
        new_link_doc = await DocumentWithLink.get(
            link_doc.id, fetch_links=True
        )
        assert new_link_doc is None

    async def test_delete_links_list(
        self, list_link_and_list_backlink_doc_pair
    ):
        link_doc, back_link_doc = list_link_and_list_backlink_doc_pair
        back_link_doc = await DocumentWithListBackLink.get(
            back_link_doc.id, fetch_links=True
        )
        await back_link_doc.delete(link_rule=DeleteRules.DELETE_LINKS)
        new_link_doc = await DocumentWithListLink.get(
            link_doc.id, fetch_links=True
        )
        assert new_link_doc is None

    async def test_init_reversed_order(self, db):
        await init_beanie(
            database=db,
            document_models=[
                DoorForReversedOrderInit,
                HouseForReversedOrderInit,
                PersonForReversedOrderInit,
            ],
        )


class TestBuildAggregations:
    async def test_find_aggregate_without_fetch_links(self, houses):
        door = await Door.find_one()
        aggregation = House.find(House.door.id == door.id).aggregate(
            [
                {"$group": {"_id": "$height", "count": {"$sum": 1}}},
            ]
        )
        assert aggregation.get_aggregation_pipeline() == [
            {"$match": {"door.$id": door.id}},
            {"$group": {"_id": "$height", "count": {"$sum": 1}}},
        ]
        result = await aggregation.to_list()
        assert result == [{"_id": 0, "count": 1}]

    async def test_find_aggregate_with_fetch_links(self, houses):
        door = await Door.find_one()
        aggregation = House.find(
            House.door.id == door.id, fetch_links=True
        ).aggregate(
            [
                {"$group": {"_id": "$height", "count": {"$sum": 1}}},
            ]
        )
        assert len(aggregation.get_aggregation_pipeline()) == 12
        assert aggregation.get_aggregation_pipeline()[10:] == [
            {"$match": {"door._id": door.id}},
            {"$group": {"_id": "$height", "count": {"$sum": 1}}},
        ]
        result = await aggregation.to_list()
        assert result == [{"_id": 0, "count": 1}]



import pytest

from beanie import init_beanie
from beanie.exceptions import MongoDBVersionError
from tests.odm.models import DocumentWithTimeseries


async def test_timeseries_collection(db):
    build_info = await db.command({"buildInfo": 1})
    mongo_version = build_info["version"]
    major_version = int(mongo_version.split(".")[0])
    if major_version < 5:
        with pytest.raises(MongoDBVersionError):
            await init_beanie(
                database=db, document_models=[DocumentWithTimeseries]
            )

    if major_version >= 5:
        await init_beanie(
            database=db, document_models=[DocumentWithTimeseries]
        )
        info = await db.command(
            {
                "listCollections": 1,
                "filter": {"name": "DocumentWithTimeseries"},
            }
        )

        assert info["cursor"]["firstBatch"][0] == {
            "name": "DocumentWithTimeseries",
            "type": "timeseries",
            "options": {
                "expireAfterSeconds": 2,
                "timeseries": {
                    "timeField": "ts",
                    "granularity": "seconds",
                    "bucketMaxSpanSeconds": 3600,
                },
            },
            "info": {"readOnly": False},
        }



import pytest
from pymongo.errors import DuplicateKeyError

from beanie.odm.fields import PydanticObjectId
from tests.odm.models import (
    DocumentTestModel,
    DocumentWithKeepNullsFalse,
    ModelWithOptionalField,
)


async def test_insert_one(document_not_inserted):
    result = await DocumentTestModel.insert_one(document_not_inserted)
    document = await DocumentTestModel.get(result.id)
    assert document is not None
    assert document.test_int == document_not_inserted.test_int
    assert document.test_list == document_not_inserted.test_list
    assert document.test_str == document_not_inserted.test_str


async def test_insert_many(documents_not_inserted):
    await DocumentTestModel.insert_many(documents_not_inserted(10))
    documents = await DocumentTestModel.find_all().to_list()
    assert len(documents) == 10


async def test_create(document_not_inserted):
    await document_not_inserted.insert()
    assert isinstance(document_not_inserted.id, PydanticObjectId)


async def test_create_twice(document_not_inserted):
    await document_not_inserted.insert()
    with pytest.raises(DuplicateKeyError):
        await document_not_inserted.insert()


async def test_insert_one_with_session(document_not_inserted, session):
    result = await DocumentTestModel.insert_one(
        document_not_inserted, session=session
    )
    document = await DocumentTestModel.get(result.id, session=session)
    assert document is not None
    assert document.test_int == document_not_inserted.test_int
    assert document.test_list == document_not_inserted.test_list
    assert document.test_str == document_not_inserted.test_str


async def test_insert_many_with_session(documents_not_inserted, session):
    await DocumentTestModel.insert_many(
        documents_not_inserted(10), session=session
    )
    documents = await DocumentTestModel.find_all(session=session).to_list()
    assert len(documents) == 10


async def test_create_with_session(document_not_inserted, session):
    await document_not_inserted.insert(session=session)
    assert isinstance(document_not_inserted.id, PydanticObjectId)


async def test_insert_keep_nulls_false():
    model = ModelWithOptionalField(i=10)
    doc = DocumentWithKeepNullsFalse(m=model)

    await doc.insert()

    new_doc = await DocumentWithKeepNullsFalse.get(doc.id)

    assert new_doc.m.i == 10
    assert new_doc.m.s is None
    assert new_doc.o is None

    raw_data = (
        await DocumentWithKeepNullsFalse.get_motor_collection().find_one(
            {"_id": doc.id}
        )
    )
    assert raw_data == {
        "_id": doc.id,
        "m": {"i": 10},
    }


async def test_insert_many_keep_nulls_false():
    models = [ModelWithOptionalField(i=10), ModelWithOptionalField(i=11)]
    docs = [DocumentWithKeepNullsFalse(m=m) for m in models]

    await DocumentWithKeepNullsFalse.insert_many(docs)

    new_docs = await DocumentWithKeepNullsFalse.find_all().to_list()

    assert len(new_docs) == 2

    assert new_docs[0].m.i == 10
    assert new_docs[0].m.s is None
    assert new_docs[0].o is None

    assert new_docs[1].m.i == 11
    assert new_docs[1].m.s is None
    assert new_docs[1].o is None

    raw_data = (
        await DocumentWithKeepNullsFalse.get_motor_collection().find_one(
            {"_id": new_docs[0].id}
        )
    )
    assert raw_data == {
        "_id": new_docs[0].id,
        "m": {"i": 10},
    }
    raw_data = (
        await DocumentWithKeepNullsFalse.get_motor_collection().find_one(
            {"_id": new_docs[1].id}
        )
    )
    assert raw_data == {
        "_id": new_docs[1].id,
        "m": {"i": 11},
    }



from decimal import Decimal

from beanie.odm.utils.pydantic import IS_PYDANTIC_V2
from tests.odm.models import DocumentWithDecimalField


def test_decimal_deserialize():
    m = DocumentWithDecimalField(amt=Decimal("1.4"))
    if IS_PYDANTIC_V2:
        m_json = m.model_dump_json()
        m_from_json = DocumentWithDecimalField.model_validate_json(m_json)
    else:
        m_json = m.json()
        m_from_json = DocumentWithDecimalField.parse_raw(m_json)
    assert isinstance(m_from_json.amt, Decimal)
    assert m_from_json.amt == Decimal("1.4")



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Dict
from typing import Mapping as MappingType

from beanie.odm.fields import (
    ExpressionField,
)

# from pydantic.fields import ModelField
# from pydantic.typing import get_origin

if TYPE_CHECKING:
    from beanie import Document


def convert_ids(
    query: MappingType[str, Any], doc: "Document", fetch_links: bool
) -> Dict[str, Any]:
    # TODO add all the cases
    new_query = {}
    for k, v in query.items():
        k_splitted = k.split(".")
        if (
            isinstance(k, ExpressionField)
            and doc.get_link_fields() is not None
            and len(k_splitted) == 2
            and k_splitted[0] in doc.get_link_fields().keys()  # type: ignore
            and k_splitted[1] == "id"
        ):
            if fetch_links:
                new_k = f"{k_splitted[0]}._id"
            else:
                new_k = f"{k_splitted[0]}.$id"
        else:
            new_k = k
        new_v: Any
        if isinstance(v, Mapping):
            new_v = convert_ids(v, doc, fetch_links)
        elif isinstance(v, list):
            new_v = [
                convert_ids(ele, doc, fetch_links)
                if isinstance(ele, Mapping)
                else ele
                for ele in v
            ]
        else:
            new_v = v

        new_query[new_k] = new_v
    return new_query



import re
from datetime import date, datetime
from enum import Enum
from uuid import uuid4

import pytest
from bson import Binary, Regex
from pydantic import AnyUrl

from beanie.odm.utils.encoder import Encoder
from beanie.odm.utils.pydantic import IS_PYDANTIC_V2
from tests.odm.models import (
    BsonRegexDoc,
    Child,
    DictEnum,
    DocumentForEncodingTest,
    DocumentForEncodingTestDate,
    DocumentWithComplexDictKey,
    DocumentWithDecimalField,
    DocumentWithEnumKeysDict,
    DocumentWithHttpUrlField,
    DocumentWithKeepNullsFalse,
    DocumentWithStringField,
    ModelWithOptionalField,
    NativeRegexDoc,
    SampleWithMutableObjects,
)


async def test_encode_datetime():
    assert isinstance(Encoder().encode(datetime.now()), datetime)

    doc = DocumentForEncodingTest(datetime_field=datetime.now())
    await doc.insert()
    new_doc = await DocumentForEncodingTest.get(doc.id)
    assert isinstance(new_doc.datetime_field, datetime)


async def test_encode_date():
    assert isinstance(Encoder().encode(datetime.now()), datetime)

    doc = DocumentForEncodingTestDate()
    await doc.insert()
    new_doc = await DocumentForEncodingTestDate.get(doc.id)
    assert new_doc.date_field == doc.date_field
    assert isinstance(new_doc.date_field, date)


async def test_encode_regex():
    raw_regex = r"^AA.*CC$"
    case_sensitive_regex = re.compile(raw_regex)
    case_insensitive_regex = re.compile(raw_regex, re.I)

    assert isinstance(Encoder().encode(case_sensitive_regex), Regex)
    assert isinstance(Encoder().encode(case_insensitive_regex), Regex)

    matching_doc = DocumentWithStringField(string_field="AABBCC")
    ignore_case_matching_doc = DocumentWithStringField(string_field="aabbcc")
    non_matching_doc = DocumentWithStringField(string_field="abc")

    for doc in (matching_doc, ignore_case_matching_doc, non_matching_doc):
        await doc.insert()

    assert {matching_doc.id, ignore_case_matching_doc.id} == {
        doc.id
        async for doc in DocumentWithStringField.find(
            DocumentWithStringField.string_field == case_insensitive_regex
        )
    }
    assert {matching_doc.id} == {
        doc.id
        async for doc in DocumentWithStringField.find(
            DocumentWithStringField.string_field == case_sensitive_regex
        )
    }


def test_encode_with_custom_encoder():
    assert isinstance(
        Encoder(custom_encoders={datetime: str}).encode(datetime.now()), str
    )


async def test_bytes():
    encoded_b = Encoder().encode(b"test")
    assert isinstance(encoded_b, Binary)
    assert encoded_b.subtype == 0

    doc = DocumentForEncodingTest(bytes_field=b"test")
    await doc.insert()
    new_doc = await DocumentForEncodingTest.get(doc.id)
    assert isinstance(new_doc.bytes_field, bytes)


async def test_bytes_already_binary():
    b = Binary(b"123", 3)
    encoded_b = Encoder().encode(b)
    assert isinstance(encoded_b, Binary)
    assert encoded_b.subtype == 3


async def test_mutable_objects_on_save():
    instance = SampleWithMutableObjects(
        d={"Bar": Child(child_field="Foo")}, lst=[Child(child_field="Bar")]
    )
    await instance.save()
    assert isinstance(instance.d["Bar"], Child)
    assert isinstance(instance.lst[0], Child)


async def test_decimal():
    test_amts = DocumentWithDecimalField(amt=1, other_amt=2)
    await test_amts.insert()
    obj = await DocumentWithDecimalField.get(test_amts.id)
    assert obj.amt == 1
    assert obj.other_amt == 2

    test_amts.amt = 6
    await test_amts.save_changes()

    obj = await DocumentWithDecimalField.get(test_amts.id)
    assert obj.amt == 6

    test_amts = (await DocumentWithDecimalField.find_all().to_list())[0]
    test_amts.other_amt = 7
    await test_amts.save_changes()

    obj = await DocumentWithDecimalField.get(test_amts.id)
    assert obj.other_amt == 7


def test_keep_nulls_false():
    model = ModelWithOptionalField(i=10)
    doc = DocumentWithKeepNullsFalse(m=model)

    encoder = Encoder(keep_nulls=False, to_db=True)
    encoded_doc = encoder.encode(doc)
    assert encoded_doc == {"m": {"i": 10}}


@pytest.mark.skipif(not IS_PYDANTIC_V2, reason="Test only for Pydantic v2")
def test_should_encode_pydantic_v2_url_correctly():
    url = AnyUrl("https://example.com")
    encoder = Encoder()
    encoded_url = encoder.encode(url)
    assert isinstance(encoded_url, str)
    # pydantic2 add trailing slash for naked url. see https://github.com/pydantic/pydantic/issues/6943
    assert encoded_url == "https://example.com/"


async def test_should_be_able_to_save_retrieve_doc_with_url():
    doc = DocumentWithHttpUrlField(url_field="https://example.com")
    assert isinstance(doc.url_field, AnyUrl)
    await doc.save()

    new_doc = await DocumentWithHttpUrlField.find_one(
        DocumentWithHttpUrlField.id == doc.id
    )

    assert isinstance(new_doc.url_field, AnyUrl)
    assert new_doc.url_field == doc.url_field


async def test_dict_with_complex_key():
    assert isinstance(Encoder().encode({uuid4(): datetime.now()}), dict)

    uuid = uuid4()
    # reset microseconds, because it looses by mongo
    dt = datetime.now().replace(microsecond=0)

    doc = DocumentWithComplexDictKey(dict_field={uuid: dt})
    await doc.insert()
    new_doc = await DocumentWithComplexDictKey.get(doc.id)

    assert isinstance(new_doc.dict_field, dict)
    assert new_doc.dict_field.get(uuid) == dt


async def test_dict_with_enum_keys():
    doc = DocumentWithEnumKeysDict(color={DictEnum.RED: "favorite"})
    await doc.save()

    assert isinstance(doc.color, dict)

    for key in doc.color:
        assert isinstance(key, Enum)
        assert key == DictEnum.RED


async def test_native_regex():
    regex = re.compile(r"^1?$|^(11+?)\1+$", (re.I | re.M | re.S) ^ re.UNICODE)
    doc = await NativeRegexDoc(regex=regex).insert()
    new_doc = await NativeRegexDoc.get(doc.id)
    assert new_doc.regex == regex
    assert new_doc.regex.pattern == r"^1?$|^(11+?)\1+$"
    assert new_doc.regex.flags == int(re.I | re.M | re.S ^ re.UNICODE)


async def test_bson_regex():
    regex = Regex(r"^1?$|^(11+?)\1+$")
    doc = await BsonRegexDoc(regex=regex).insert()
    new_doc = await BsonRegexDoc.get(doc.id)
    assert new_doc.regex == Regex(pattern=r"^1?$|^(11+?)\1+$")



from pathlib import Path

import toml

from beanie import __version__


def parse_version_from_pyproject():
    pyproject = Path(__file__).parent.parent / "pyproject.toml"
    toml_data = toml.loads(pyproject.read_text())
    return toml_data["project"]["version"]


def test_version():
    assert __version__ == parse_version_from_pyproject()



import subprocess
from dataclasses import dataclass
from datetime import datetime
from typing import List

import requests  # type: ignore


@dataclass
class PullRequest:
    number: int
    title: str
    user: str
    user_url: str
    url: str


class GitHubHandler:
    def __init__(
        self,
        username: str,
        repository: str,
        current_version: str,
        new_version: str,
    ):
        self.username = username
        self.repository = repository
        self.base_url = f"https://api.github.com/repos/{username}/{repository}"
        self.current_version = current_version
        self.new_version = new_version
        self.commits = self.get_commits_after_tag(current_version)
        self.prs = [self.get_pr_for_commit(commit) for commit in self.commits]

    def get_commits_after_tag(self, tag: str) -> List[str]:
        result = subprocess.run(
            ["git", "log", f"{tag}..HEAD", "--pretty=format:%H"],
            stdout=subprocess.PIPE,
            text=True,
        )
        return result.stdout.split()

    def get_pr_for_commit(self, commit_sha: str) -> PullRequest:
        url = f"{self.base_url}/commits/{commit_sha}/pulls"
        response = requests.get(url)
        response.raise_for_status()
        pr_data = response.json()[0]
        return PullRequest(
            number=pr_data["number"],
            title=pr_data["title"],
            user=pr_data["user"]["login"],
            user_url=pr_data["user"]["html_url"],
            url=pr_data["html_url"],
        )

    def build_markdown_for_many_prs(self) -> str:
        markdown = f"\n## [{self.new_version}] - {datetime.now().strftime('%Y-%m-%d')}\n"
        for pr in self.prs:
            markdown += (
                f"### {pr.title.capitalize()}\n"
                f"- Author - [{pr.user}]({pr.user_url})\n"
                f"- PR <{pr.url}>\n"
            )
        markdown += f"\n[{self.new_version}]: https://pypi.org/project/{self.repository}/{self.new_version}\n"
        return markdown

    def commit_changes(self):
        self.run_git_command(
            ["git", "config", "--global", "user.name", "github-actions[bot]"]
        )
        self.run_git_command(
            [
                "git",
                "config",
                "--global",
                "user.email",
                "github-actions[bot]@users.noreply.github.com",
            ]
        )
        self.run_git_command(["git", "add", "."])
        self.run_git_command(
            ["git", "commit", "-m", f"Bump version to {self.new_version}"]
        )
        self.run_git_command(["git", "tag", self.new_version])
        self.git_push()

    def git_push(self):
        self.run_git_command(["git", "push", "origin", "main", "--tags"])

    @staticmethod
    def run_git_command(command: List[str]):
        subprocess.run(command, check=True)



from pydantic.main import BaseModel

from beanie import Document, iterative_migration


class Tag(BaseModel):
    color: str
    name: str


class Note(Document):
    title: str
    tag: Tag

    class Settings:
        name = "notes"


class Forward:
    @iterative_migration()
    async def change_title_6(
        self, input_document: Note, output_document: Note
    ):
        if input_document.title == "6":
            output_document.title = "six"

    @iterative_migration()
    async def change_title_7(
        self, input_document: Note, output_document: Note
    ):
        if input_document.title == "7":
            output_document.title = "seven"


class Backward:
    @iterative_migration()
    async def change_title_7(
        self, input_document: Note, output_document: Note
    ):
        if input_document.title == "seven":
            output_document.title = "7"

    @iterative_migration()
    async def change_title_6(
        self, input_document: Note, output_document: Note
    ):
        if input_document.title == "six":
            output_document.title = "6"



import re

import bson
import pydantic
from typing_extensions import Annotated

from beanie.odm.utils.pydantic import IS_PYDANTIC_V2


def _to_bson_regex(v):
    return v.try_compile() if isinstance(v, bson.Regex) else v


if IS_PYDANTIC_V2:
    Pattern = Annotated[
        re.Pattern,
        pydantic.BeforeValidator(
            lambda v: v.try_compile() if isinstance(v, bson.Regex) else v
        ),
    ]
else:

    class Pattern(bson.Regex):  # type: ignore[no-redef]
        @classmethod
        def __get_validators__(cls):
            yield _to_bson_regex



from beanie.odm.fields import Link
from beanie.odm.views import View
from tests.odm.models import DocumentTestModel, DocumentTestModelWithLink


class ViewForTest(View):
    number: int
    string: str

    class Settings:
        view_name = "test_view"
        source = DocumentTestModel
        pipeline = [
            {"$match": {"$expr": {"$gt": ["$test_int", 8]}}},
            {"$project": {"number": "$test_int", "string": "$test_str"}},
        ]


class ViewForTestWithLink(View):
    link: Link[DocumentTestModel]

    class Settings:
        view_name = "test_view_with_link"
        source = DocumentTestModelWithLink
        pipeline = [{"$set": {"link": "$test_link"}}]



from typing import TYPE_CHECKING, Any, Dict, Type, Union

from pydantic import BaseModel

from beanie.exceptions import (
    ApplyChangesException,
    DocWasNotRegisteredInUnionClass,
    UnionHasNoRegisteredDocs,
)
from beanie.odm.interfaces.detector import ModelType
from beanie.odm.utils.pydantic import get_config_value, parse_model

if TYPE_CHECKING:
    from beanie.odm.documents import Document


def merge_models(left: BaseModel, right: BaseModel) -> None:
    """
    Merge two models
    :param left: left model
    :param right: right model
    :return: None
    """
    from beanie.odm.fields import Link

    for k, right_value in right.__iter__():
        left_value = getattr(left, k)
        if isinstance(right_value, BaseModel) and isinstance(
            left_value, BaseModel
        ):
            if get_config_value(left_value, "frozen"):
                left.__setattr__(k, right_value)
            else:
                merge_models(left_value, right_value)
            continue
        if isinstance(right_value, list):
            links_found = False
            for i in right_value:
                if isinstance(i, Link):
                    links_found = True
                    break
            if links_found:
                continue
            left.__setattr__(k, right_value)
        elif not isinstance(right_value, Link):
            left.__setattr__(k, right_value)


def apply_changes(
    changes: Dict[str, Any], target: Union[BaseModel, Dict[str, Any]]
):
    for key, value in changes.items():
        if "." in key:
            key_parts = key.split(".")
            current_target = target
            try:
                for part in key_parts[:-1]:
                    if isinstance(current_target, dict):
                        current_target = current_target[part]
                    elif isinstance(current_target, BaseModel):
                        current_target = getattr(current_target, part)
                    else:
                        raise ApplyChangesException(
                            f"Unexpected type of target: {type(target)}"
                        )
                final_key = key_parts[-1]
                if isinstance(current_target, dict):
                    current_target[final_key] = value
                elif isinstance(current_target, BaseModel):
                    setattr(current_target, final_key, value)
                else:
                    raise ApplyChangesException(
                        f"Unexpected type of target: {type(target)}"
                    )
            except (KeyError, AttributeError) as e:
                raise ApplyChangesException(
                    f"Failed to apply change for key '{key}': {e}"
                )
        else:
            if isinstance(target, dict):
                target[key] = value
            elif isinstance(target, BaseModel):
                setattr(target, key, value)
            else:
                raise ApplyChangesException(
                    f"Unexpected type of target: {type(target)}"
                )


def save_state(item: BaseModel):
    if hasattr(item, "_save_state"):
        item._save_state()  # type: ignore


def parse_obj(
    model: Union[Type[BaseModel], Type["Document"]],
    data: Any,
    lazy_parse: bool = False,
) -> BaseModel:
    if (
        hasattr(model, "get_model_type")
        and model.get_model_type() == ModelType.UnionDoc  # type: ignore
    ):
        if model._document_models is None:  # type: ignore
            raise UnionHasNoRegisteredDocs

        if isinstance(data, dict):
            class_name = data[model.get_settings().class_id]  # type: ignore
        else:
            class_name = data._class_id

        if class_name not in model._document_models:  # type: ignore
            raise DocWasNotRegisteredInUnionClass
        return parse_obj(
            model=model._document_models[class_name],  # type: ignore
            data=data,
            lazy_parse=lazy_parse,
        )  # type: ignore
    if (
        hasattr(model, "get_model_type")
        and model.get_model_type() == ModelType.Document  # type: ignore
        and model._inheritance_inited  # type: ignore
    ):
        if isinstance(data, dict):
            class_name = data.get(model.get_settings().class_id)  # type: ignore
        elif hasattr(data, model.get_settings().class_id):  # type: ignore
            class_name = data._class_id
        else:
            class_name = None

        if model._children and class_name in model._children:  # type: ignore
            return parse_obj(
                model=model._children[class_name],  # type: ignore
                data=data,
                lazy_parse=lazy_parse,
            )  # type: ignore

    if (
        lazy_parse
        and hasattr(model, "get_model_type")
        and model.get_model_type() == ModelType.Document  # type: ignore
    ):
        o = model.lazy_parse(data, {"_id"})  # type: ignore
        o._saved_state = {"_id": o.id}
        return o
    result = parse_model(model, data)
    save_state(result)
    return result



from pydantic import Field
from pydantic.main import BaseModel

from tests.odm.models import DocumentTestModel


async def test_aggregate(documents):
    await documents(4, "uno")
    await documents(2, "dos")
    await documents(1, "cuatro")
    result = await DocumentTestModel.aggregate(
        [{"$group": {"_id": "$test_str", "total": {"$sum": "$test_int"}}}]
    ).to_list()
    assert len(result) == 3
    assert {"_id": "cuatro", "total": 0} in result
    assert {"_id": "dos", "total": 1} in result
    assert {"_id": "uno", "total": 6} in result


async def test_aggregate_with_filter(documents):
    await documents(4, "uno")
    await documents(2, "dos")
    await documents(1, "cuatro")
    result = (
        await DocumentTestModel.find(DocumentTestModel.test_int >= 1)
        .aggregate(
            [{"$group": {"_id": "$test_str", "total": {"$sum": "$test_int"}}}]
        )
        .to_list()
    )
    assert len(result) == 2
    assert {"_id": "dos", "total": 1} in result
    assert {"_id": "uno", "total": 6} in result


async def test_aggregate_with_item_model(documents):
    class OutputItem(BaseModel):
        id: str = Field(None, alias="_id")
        total: int

    await documents(4, "uno")
    await documents(2, "dos")
    await documents(1, "cuatro")
    ids = []
    async for i in DocumentTestModel.aggregate(
        [{"$group": {"_id": "$test_str", "total": {"$sum": "$test_int"}}}],
        projection_model=OutputItem,
    ):
        if i.id == "cuatro":
            assert i.total == 0
        elif i.id == "dos":
            assert i.total == 1
        elif i.id == "uno":
            assert i.total == 6
        else:
            raise KeyError
        ids.append(i.id)
    assert set(ids) == {"cuatro", "dos", "uno"}


async def test_aggregate_with_session(documents, session):
    await documents(4, "uno")
    await documents(2, "dos")
    await documents(1, "cuatro")
    result = await DocumentTestModel.aggregate(
        [{"$group": {"_id": "$test_str", "total": {"$sum": "$test_int"}}}],
        session=session,
    ).to_list()
    assert len(result) == 3
    assert {"_id": "cuatro", "total": 0} in result
    assert {"_id": "dos", "total": 1} in result
    assert {"_id": "uno", "total": 6} in result



from pydantic.main import BaseModel

from beanie import Document, Indexed, PydanticObjectId, iterative_migration


class Tag(BaseModel):
    color: str
    name: str


class OldNote(Document):
    name: Indexed(str, unique=True)
    tag: Tag

    class Settings:
        name = "notes"


class Note(Document):
    name: Indexed(str, unique=True)
    title: str
    tag: Tag

    class Settings:
        name = "notes"


fixed_id = PydanticObjectId("6076f1f3e4b7f6b7a0f6e5a0")


class Forward:
    @iterative_migration(batch_size=2)
    async def name_to_title(
        self, input_document: OldNote, output_document: Note
    ):
        output_document.title = input_document.name
        if output_document.title > "5":
            output_document.name = "5"


class Backward:
    @iterative_migration()
    async def title_to_name(
        self, input_document: Note, output_document: OldNote
    ):
        output_document.name = input_document.title



import pytest
from pydantic import ValidationError

from tests.odm.models import DocumentWithPydanticConfig


def test_pydantic_config():
    doc = DocumentWithPydanticConfig(num_1=2)
    with pytest.raises(ValidationError):
        doc.num_1 = "wrong"



from typing import Optional

import pytest
from bson import ObjectId
from pydantic import BaseModel, ValidationError

from beanie import PydanticObjectId
from beanie.odm.utils.pydantic import IS_PYDANTIC_V2
from tests.odm.models import (
    DocumentWithValidationOnSave,
    Lock,
    WindowWithValidationOnSave,
)


async def test_validate_on_insert():
    doc = DocumentWithValidationOnSave(num_1=1, num_2=2)
    doc.num_1 = "wrong_value"
    with pytest.raises(ValidationError):
        await doc.insert()


async def test_validate_on_replace():
    doc = DocumentWithValidationOnSave(num_1=1, num_2=2)
    await doc.insert()
    doc.num_1 = "wrong_value"
    with pytest.raises(ValidationError):
        await doc.replace()


async def test_validate_on_save_changes():
    doc = DocumentWithValidationOnSave(num_1=1, num_2=2)
    await doc.insert()
    doc.num_1 = "wrong_value"
    with pytest.raises(ValidationError):
        await doc.save_changes()


async def test_validate_on_save_keep_the_id_type():
    class UpdateModel(BaseModel):
        num_1: Optional[int] = None
        related: Optional[PydanticObjectId] = None

    doc = DocumentWithValidationOnSave(num_1=1, num_2=2)
    await doc.insert()
    update = UpdateModel(related=PydanticObjectId())
    if IS_PYDANTIC_V2:
        doc = doc.model_copy(update=update.model_dump(exclude_unset=True))
    else:
        doc = doc.copy(update=update.dict(exclude_unset=True))
    doc.num_2 = 1000
    await doc.save()
    in_db = await DocumentWithValidationOnSave.get_motor_collection().find_one(
        {"_id": doc.id}
    )
    assert isinstance(in_db["related"], ObjectId)
    new_doc = await DocumentWithValidationOnSave.get(doc.id)
    assert isinstance(new_doc.related, PydanticObjectId)


async def test_validate_on_save_action():
    doc = DocumentWithValidationOnSave(num_1=1, num_2=2)
    await doc.insert()
    assert doc.num_2 == 3


async def test_validate_on_save_skip_action():
    doc = DocumentWithValidationOnSave(num_1=1, num_2=2)
    await doc.insert(skip_actions=["num_2_plus_1"])
    assert doc.num_2 == 2


async def test_validate_on_save_dbref():
    lock = Lock(k=1)
    await lock.insert()
    window = WindowWithValidationOnSave(
        x=1,
        y=1,
        lock=lock.to_ref(),  # this is what exactly we want to test
    )
    await window.insert()



from __future__ import annotations

import asyncio
from collections import OrderedDict
from dataclasses import dataclass
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from typing import OrderedDict as OrderedDictType

from bson import DBRef, ObjectId
from bson.errors import InvalidId
from pydantic import BaseModel
from pymongo import ASCENDING, IndexModel
from typing_extensions import get_args

from beanie.odm.enums import SortDirection
from beanie.odm.operators.find.comparison import (
    GT,
    GTE,
    LT,
    LTE,
    NE,
    Eq,
    In,
)
from beanie.odm.registry import DocsRegistry
from beanie.odm.utils.parsing import parse_obj
from beanie.odm.utils.pydantic import (
    IS_PYDANTIC_V2,
    IS_PYDANTIC_V2_10,
    get_field_type,
    get_model_fields,
    parse_object_as,
)

if IS_PYDANTIC_V2:
    from pydantic import (
        GetCoreSchemaHandler,
        GetJsonSchemaHandler,
        TypeAdapter,
    )
    from pydantic.json_schema import JsonSchemaValue
    from pydantic_core.core_schema import (
        CoreSchema,
        ValidationInfo,
        any_schema,
        dict_schema,
        json_or_python_schema,
        no_info_after_validator_function,
        no_info_plain_validator_function,
        plain_serializer_function_ser_schema,
        simple_ser_schema,
        str_schema,
        typed_dict_field,
        typed_dict_schema,
        union_schema,
        with_info_plain_validator_function,
    )
else:
    from pydantic.fields import ModelField
    from pydantic.json import ENCODERS_BY_TYPE

if TYPE_CHECKING:
    from beanie.odm.documents import DocType


@dataclass(frozen=True)
class IndexedAnnotation:
    _indexed: Tuple[int, Dict[str, Any]]


def Indexed(typ=None, index_type=ASCENDING, **kwargs: Any):
    """
    If `typ` is defined, returns a subclass of `typ` with an extra attribute
    `_indexed` as a tuple:
    - Index 0: `index_type` such as `pymongo.ASCENDING`
    - Index 1: `kwargs` passed to `IndexModel`
    When instantiated the type of the result will actually be `typ`.

    When `typ` is not defined, returns an `IndexedAnnotation` instance, to be
    used as metadata in `Annotated` fields.

    Example:
    ```py
    # Both fields would have the same behavior
    class MyModel(BaseModel):
        field1: Indexed(str, unique=True)
        field2: Annotated[str, Indexed(unique=True)]
    ```
    """
    if typ is None:
        return IndexedAnnotation(_indexed=(index_type, kwargs))

    class NewType(typ):
        _indexed = (index_type, kwargs)

        def __new__(cls, *args: Any, **kwargs: Any):
            return typ.__new__(typ, *args, **kwargs)

        if IS_PYDANTIC_V2:

            @classmethod
            def __get_pydantic_core_schema__(
                cls, _source_type: Type[Any], _handler: GetCoreSchemaHandler
            ) -> CoreSchema:
                custom_type = getattr(
                    typ, "__get_pydantic_core_schema__", None
                )
                if custom_type is not None:
                    return custom_type(_source_type, _handler)

                return no_info_after_validator_function(
                    lambda v: v, simple_ser_schema(typ.__name__)
                )

    NewType.__name__ = f"Indexed {typ.__name__}"
    return NewType


class PydanticObjectId(ObjectId):
    """
    Object Id field. Compatible with Pydantic.
    """

    @classmethod
    def _validate(cls, v):
        if isinstance(v, bytes):
            v = v.decode("utf-8")
        try:
            return PydanticObjectId(v)
        except (InvalidId, TypeError):
            raise ValueError("Id must be of type PydanticObjectId")

    if IS_PYDANTIC_V2:

        @classmethod
        def __get_pydantic_core_schema__(
            cls, source_type: Type[Any], handler: GetCoreSchemaHandler
        ) -> CoreSchema:
            if not IS_PYDANTIC_V2_10:
                return no_info_plain_validator_function(
                    cls._validate,
                    metadata={
                        "pydantic_js_input_core_schema": str_schema(
                            pattern="^[0-9a-f]{24}$",
                            min_length=24,
                            max_length=24,
                        )
                    },
                    serialization=plain_serializer_function_ser_schema(
                        lambda instance: str(instance),
                        return_schema=str_schema(),
                        when_used="json",
                    ),
                )
            return no_info_plain_validator_function(
                cls._validate,
                json_schema_input_schema=str_schema(
                    pattern="^[0-9a-f]{24}$",
                    min_length=24,
                    max_length=24,
                ),
                serialization=plain_serializer_function_ser_schema(
                    lambda instance: str(instance),
                    return_schema=str_schema(),
                    when_used="json",
                ),
            )

        @classmethod
        def __get_pydantic_json_schema__(
            cls, schema: CoreSchema, handler: GetJsonSchemaHandler
        ) -> JsonSchemaValue:
            json_schema = handler(schema)
            json_schema.update(
                type="string",
                example="5eb7cf5a86d9755df3a6c593",
            )
            return json_schema

    else:

        @classmethod
        def __get_validators__(cls):
            yield cls._validate

        @classmethod
        def __modify_schema__(cls, field_schema: Dict[str, Any]):
            field_schema.update(
                type="string",
                example="5eb7cf5a86d9755df3a6c593",
            )


if not IS_PYDANTIC_V2:
    ENCODERS_BY_TYPE[PydanticObjectId] = (
        str  # it is a workaround to force pydantic make json schema for this field
    )

BeanieObjectId = PydanticObjectId


class ExpressionField(str):
    def __getitem__(self, item):
        """
        Get sub field

        :param item: name of the subfield
        :return: ExpressionField
        """
        return ExpressionField(f"{self}.{item}")

    def __getattr__(self, item):
        """
        Get sub field

        :param item: name of the subfield
        :return: ExpressionField
        """
        return ExpressionField(f"{self}.{item}")

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if isinstance(other, ExpressionField):
            return super(ExpressionField, self).__eq__(other)
        return Eq(field=self, other=other)

    def __gt__(self, other):
        return GT(field=self, other=other)

    def __ge__(self, other):
        return GTE(field=self, other=other)

    def __lt__(self, other):
        return LT(field=self, other=other)

    def __le__(self, other):
        return LTE(field=self, other=other)

    def __ne__(self, other):
        return NE(field=self, other=other)

    def __pos__(self):
        return self, SortDirection.ASCENDING

    def __neg__(self):
        return self, SortDirection.DESCENDING

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self


class DeleteRules(str, Enum):
    DO_NOTHING = "DO_NOTHING"
    DELETE_LINKS = "DELETE_LINKS"


class WriteRules(str, Enum):
    DO_NOTHING = "DO_NOTHING"
    WRITE = "WRITE"


class LinkTypes(str, Enum):
    DIRECT = "DIRECT"
    OPTIONAL_DIRECT = "OPTIONAL_DIRECT"
    LIST = "LIST"
    OPTIONAL_LIST = "OPTIONAL_LIST"

    BACK_DIRECT = "BACK_DIRECT"
    BACK_LIST = "BACK_LIST"
    OPTIONAL_BACK_DIRECT = "OPTIONAL_BACK_DIRECT"
    OPTIONAL_BACK_LIST = "OPTIONAL_BACK_LIST"


class LinkInfo(BaseModel):
    field_name: str
    lookup_field_name: str
    document_class: Type[BaseModel]  # Document class
    link_type: LinkTypes
    nested_links: Optional[Dict] = None
    is_fetchable: bool = True


T = TypeVar("T")


class Link(Generic[T]):
    def __init__(self, ref: DBRef, document_class: Type[T]):
        self.ref = ref
        self.document_class = document_class

    async def fetch(self, fetch_links: bool = False) -> Union[T, Link[T]]:
        result = await self.document_class.get(  # type: ignore
            self.ref.id, with_children=True, fetch_links=fetch_links
        )
        return result or self

    @classmethod
    async def fetch_one(cls, link: Link[T]):
        return await link.fetch()

    @classmethod
    async def fetch_list(
        cls,
        links: List[Union[Link[T], DocType]],
        fetch_links: bool = False,
    ):
        """
        Fetch list that contains links and documents
        :param links:
        :param fetch_links:
        :return:
        """
        data = Link.repack_links(links)  # type: ignore
        ids_to_fetch = []
        document_class = None
        for doc_id, link in data.items():
            if isinstance(link, Link):
                if document_class is None:
                    document_class = link.document_class
                else:
                    if document_class != link.document_class:
                        raise ValueError(
                            "All the links must have the same model class"
                        )
                ids_to_fetch.append(link.ref.id)

        if ids_to_fetch:
            fetched_models = await document_class.find(  # type: ignore
                In("_id", ids_to_fetch),
                with_children=True,
                fetch_links=fetch_links,
            ).to_list()

            for model in fetched_models:
                data[model.id] = model

        return list(data.values())

    @staticmethod
    def repack_links(
        links: List[Union[Link[T], DocType]],
    ) -> OrderedDictType[Any, Any]:
        result = OrderedDict()
        for link in links:
            if isinstance(link, Link):
                result[link.ref.id] = link
            else:
                result[link.id] = link
        return result

    @classmethod
    async def fetch_many(cls, links: List[Link[T]]) -> List[Union[T, Link[T]]]:
        coros = []
        for link in links:
            coros.append(link.fetch())
        return await asyncio.gather(*coros)

    if IS_PYDANTIC_V2:

        @staticmethod
        def serialize(value: Union[Link[T], BaseModel]):
            if isinstance(value, Link):
                return value.to_dict()
            return value.model_dump(mode="json")

        @classmethod
        def wrapped_validate(
            cls, source_type: Type[Any], handler: GetCoreSchemaHandler
        ):
            def validate(
                v: Union[Link[T], T, DBRef, dict[str, Any]],
                validation_info: ValidationInfo,
            ) -> Link[T] | T:
                document_class = DocsRegistry.evaluate_fr(  # type: ignore
                    get_args(source_type)[0]
                )

                if isinstance(v, DBRef):
                    return cls(ref=v, document_class=document_class)
                if isinstance(v, Link):
                    return v
                if isinstance(v, dict) and v.keys() == {"id", "collection"}:
                    return cls(
                        ref=DBRef(
                            collection=v["collection"],
                            id=TypeAdapter(
                                document_class.model_fields["id"].annotation
                            ).validate_python(v["id"]),
                        ),
                        document_class=document_class,
                    )
                if isinstance(v, dict) or isinstance(v, BaseModel):
                    return parse_obj(document_class, v)

                # Default fallback case for unknown type
                new_id = TypeAdapter(
                    document_class.model_fields["id"].annotation
                ).validate_python(v)
                ref = DBRef(
                    collection=document_class.get_collection_name(), id=new_id
                )
                return cls(ref=ref, document_class=document_class)

            return validate

        @classmethod
        def __get_pydantic_core_schema__(
            cls, source_type: Type[Any], handler: GetCoreSchemaHandler
        ) -> CoreSchema:
            return json_or_python_schema(
                python_schema=with_info_plain_validator_function(
                    cls.wrapped_validate(source_type, handler)
                ),
                json_schema=union_schema(
                    [
                        typed_dict_schema(
                            {
                                "id": typed_dict_field(str_schema()),
                                "collection": typed_dict_field(str_schema()),
                            }
                        ),
                        dict_schema(
                            keys_schema=str_schema(),
                            values_schema=any_schema(),
                        ),
                    ]
                ),
                serialization=plain_serializer_function_ser_schema(
                    function=lambda instance: cls.serialize(instance),
                    when_used="json-unless-none",
                ),
            )

    else:

        @classmethod
        def __get_validators__(cls):
            yield cls._validate

        @classmethod
        def _validate(
            cls,
            v: Union[Link[T], T, DBRef, dict[str, Any]],
            field: ModelField,
        ) -> Link[T] | T:
            document_class = DocsRegistry.evaluate_fr(  # type: ignore
                field.sub_fields[0].type_
            )

            if isinstance(v, DBRef):
                return cls(ref=v, document_class=document_class)
            if isinstance(v, Link):
                return v
            if isinstance(v, dict) or isinstance(v, BaseModel):
                return parse_obj(document_class, v)

            # Default fallback case for unknown type
            new_id = parse_object_as(
                get_field_type(get_model_fields(document_class)["id"]), v
            )
            ref = DBRef(
                collection=document_class.get_collection_name(), id=new_id
            )
            return cls(ref=ref, document_class=document_class)

        @classmethod
        def __modify_schema__(cls, field_schema: Dict[str, Any]):
            field_schema.clear()
            field_schema.update(
                {
                    "anyOf": [
                        {
                            "properties": {
                                "id": {"type": "string", "title": "Id"},
                                "collection": {
                                    "type": "string",
                                    "title": "Collection",
                                },
                            },
                            "type": "object",
                            "required": ["id", "collection"],
                        },
                        {"type": "object"},
                    ],
                }
            )

    def to_ref(self):
        return self.ref

    def to_dict(self):
        return {"id": str(self.ref.id), "collection": self.ref.collection}


if not IS_PYDANTIC_V2:
    ENCODERS_BY_TYPE[Link] = lambda o: o.to_dict()


class BackLink(Generic[T]):
    """Back reference to a document"""

    def __init__(self, document_class: Type[T]):
        self.document_class = document_class

    if IS_PYDANTIC_V2:

        @classmethod
        def wrapped_validate(
            cls, source_type: Type[Any], handler: GetCoreSchemaHandler
        ):
            def validate(
                v: Union[T, dict[str, Any]], validation_info: ValidationInfo
            ) -> BackLink[T] | T:
                document_class = DocsRegistry.evaluate_fr(  # type: ignore
                    get_args(source_type)[0]
                )
                if isinstance(v, dict) or isinstance(v, BaseModel):
                    return parse_obj(document_class, v)
                return cls(document_class=document_class)

            return validate

        @classmethod
        def __get_pydantic_core_schema__(
            cls, source_type: Type[Any], handler: GetCoreSchemaHandler
        ) -> CoreSchema:
            # NOTE: BackLinks are only virtual fields, they shouldn't be serialized nor appear in the schema.
            return json_or_python_schema(
                python_schema=with_info_plain_validator_function(
                    cls.wrapped_validate(source_type, handler)
                ),
                json_schema=dict_schema(
                    keys_schema=str_schema(),
                    values_schema=any_schema(),
                ),
                serialization=plain_serializer_function_ser_schema(
                    lambda instance: cls.to_dict(instance),
                    return_schema=dict_schema(),
                    when_used="json-unless-none",
                ),
            )

    else:

        @classmethod
        def __get_validators__(cls):
            yield cls._validate

        @classmethod
        def _validate(
            cls, v: Union[T, dict[str, Any]], field: ModelField
        ) -> BackLink[T] | T:
            document_class = DocsRegistry.evaluate_fr(  # type: ignore
                field.sub_fields[0].type_
            )
            if isinstance(v, dict) or isinstance(v, BaseModel):
                return parse_obj(document_class, v)
            return cls(document_class=document_class)

        @classmethod
        def __modify_schema__(cls, field_schema: Dict[str, Any]):
            field_schema.clear()
            field_schema.update(
                {
                    "anyOf": [
                        {
                            "properties": {
                                "id": {"type": "string", "title": "Id"},
                                "collection": {
                                    "type": "string",
                                    "title": "Collection",
                                },
                            },
                            "type": "object",
                            "required": ["id", "collection"],
                        },
                        {"type": "object"},
                    ],
                }
            )

    def to_dict(self) -> dict[str, str]:
        document_class = DocsRegistry.evaluate_fr(self.document_class)  # type: ignore
        return {"collection": document_class.get_collection_name()}


if not IS_PYDANTIC_V2:
    ENCODERS_BY_TYPE[BackLink] = lambda o: o.to_dict()


class IndexModelField:
    def __init__(self, index: IndexModel):
        self.index = index
        self.name = index.document["name"]

        self.fields = tuple(sorted(self.index.document["key"]))
        self.options = tuple(
            sorted(
                (k, v)
                for k, v in self.index.document.items()
                if k not in ["key", "v"]
            )
        )

    def __eq__(self, other):
        return self.fields == other.fields and self.options == other.options

    def __repr__(self):
        return f"IndexModelField({self.name}, {self.fields}, {self.options})"

    @staticmethod
    def list_difference(
        left: List[IndexModelField], right: List[IndexModelField]
    ):
        result = []
        for index in left:
            if index not in right:
                result.append(index)
        return result

    @staticmethod
    def list_to_index_model(left: List[IndexModelField]):
        return [index.index for index in left]

    @classmethod
    def from_motor_index_information(cls, index_info: dict):
        result = []
        for name, details in index_info.items():
            fields = details["key"]
            if ("_id", 1) in fields:
                continue

            options = {k: v for k, v in details.items() if k != "key"}
            index_model = IndexModelField(
                IndexModel(fields, name=name, **options)
            )
            result.append(index_model)
        return result

    def same_fields(self, other: IndexModelField):
        return self.fields == other.fields

    @staticmethod
    def find_index_with_the_same_fields(
        indexes: List[IndexModelField], index: IndexModelField
    ):
        for i in indexes:
            if i.same_fields(index):
                return i
        return None

    @staticmethod
    def merge_indexes(
        left: List[IndexModelField], right: List[IndexModelField]
    ):
        left_dict = {index.fields: index for index in left}
        right_dict = {index.fields: index for index in right}
        left_dict.update(right_dict)
        return list(left_dict.values())

    @classmethod
    def _validate(cls, v: Any) -> "IndexModelField":
        if isinstance(v, IndexModel):
            return IndexModelField(v)
        else:
            return IndexModelField(IndexModel(v))

    if IS_PYDANTIC_V2:

        @classmethod
        def __get_pydantic_core_schema__(
            cls, source_type: Type[Any], handler: GetCoreSchemaHandler
        ) -> CoreSchema:
            return no_info_plain_validator_function(cls._validate)

    else:

        @classmethod
        def __get_validators__(cls):
            yield cls._validate



from typing import List, Optional

from pydantic import Field

from beanie.odm.fields import IndexModelField
from beanie.odm.settings.base import ItemSettings
from beanie.odm.settings.timeseries import TimeSeriesConfig
from beanie.odm.utils.pydantic import IS_PYDANTIC_V2

if IS_PYDANTIC_V2:
    from pydantic import ConfigDict


class DocumentSettings(ItemSettings):
    use_state_management: bool = False
    state_management_replace_objects: bool = False
    state_management_save_previous: bool = False
    validate_on_save: bool = False
    use_revision: bool = False
    single_root_inheritance: bool = False

    indexes: List[IndexModelField] = Field(default_factory=list)
    merge_indexes: bool = False
    timeseries: Optional[TimeSeriesConfig] = None

    lazy_parsing: bool = False

    keep_nulls: bool = True

    max_nesting_depths_per_field: dict = Field(default_factory=dict)
    max_nesting_depth: int = 3

    if IS_PYDANTIC_V2:
        model_config = ConfigDict(
            arbitrary_types_allowed=True,
        )
    else:

        class Config:
            arbitrary_types_allowed = True



import pytest

from beanie.exceptions import ApplyChangesException
from beanie.odm.documents import MergeStrategy
from tests.odm.models import DocumentToTestSync


class TestSync:
    async def test_merge_remote(self):
        doc = DocumentToTestSync()
        await doc.insert()

        doc2 = await DocumentToTestSync.get(doc.id)
        doc2.s = "foo"

        doc.i = 100
        await doc.save()

        await doc2.sync()

        assert doc2.s == "TEST"
        assert doc2.i == 100

    async def test_merge_local(self):
        doc = DocumentToTestSync(d={"option_1": {"s": "foo"}})
        await doc.insert()

        doc2 = await DocumentToTestSync.get(doc.id)
        doc2.s = "foo"
        doc2.n.option_1.s = "bar"
        doc2.d["option_1"]["s"] = "bar"

        doc.i = 100
        await doc.save()

        await doc2.sync(merge_strategy=MergeStrategy.local)

        assert doc2.s == "foo"
        assert doc2.n.option_1.s == "bar"
        assert doc2.d["option_1"]["s"] == "bar"

        assert doc2.i == 100

    async def test_merge_local_impossible_apply_changes(self):
        doc = DocumentToTestSync(d={"option_1": {"s": "foo"}})
        await doc.insert()

        doc2 = await DocumentToTestSync.get(doc.id)
        doc2.d["option_1"]["s"] = {"foo": "bar"}

        doc.d = {"option_1": "nothing"}
        await doc.save()
        with pytest.raises(ApplyChangesException):
            await doc2.sync(merge_strategy=MergeStrategy.local)
